# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the RETIS method(s)."""
import copy
import logging
import os
import tempfile
import unittest
from io import StringIO
from unittest.mock import patch
from pyretis.core.common import (relative_shoots_select,
                                 null_move, soft_partial_exit,
                                 compute_weight)
from pyretis.core.pathensemble import PathEnsemble
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.core.retis import (high_acc_swap,
                                make_retis_step,
                                repptis_swap,
                                retis_moves,
                                retis_swap,
                                retis_swap_wrapper,
                                retis_swap_zero)
from pyretis.inout.common import make_dirs
from pyretis.initiation import initiate_path_simulation
from pyretis.inout.settings import fill_up_tis_and_retis_settings
from .help import (
    create_ensembles_and_paths,
    make_internal_path,
    MockEngine,
    MockOrder,
    prepare_test_simulation,
)

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


def compare_path_skip_generated(path1, path2):
    """Compare two paths, but skip the generated attribute."""
    gen1 = copy.deepcopy(path1.generated)
    gen2 = copy.deepcopy(path2.generated)
    del path1.generated
    del path2.generated
    equal = path1 == path2
    path1.generated = gen1
    path2.generated = gen2
    return equal


class RetisTestSwap(unittest.TestCase):
    """The RETIS specific methods, for Internal simulations."""

    def test_swap_internal(self):
        """Test swapping of paths."""
        settings, ensembles = create_ensembles_and_paths()
        # 1) Try [0^+] with [1^+]:
        # This move should be rejected, we here check that
        # the currently accepted paths are not modified.
        path1 = ensembles[1]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[2]['path_ensemble'].last_path
        path2c = path2.copy()

        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '001'))
            make_dirs(os.path.join(tempdir, '002'))
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            settings['ensemble'][2]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=1, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'NCR')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertTrue(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertTrue(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

        # 2) Try [3^+] with [4^+]
        # This move should be accepted:
        path1 = ensembles[3]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[4]['path_ensemble'].last_path
        path2c = path2.copy()

        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '003'))
            make_dirs(os.path.join(tempdir, '004'))
            settings['ensemble'][3]['simulation']['exe_path'] = tempdir
            settings['ensemble'][4]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=3, settings=settings, cycle=0)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Here, path1 and trial2 should be identical:
        self.assertTrue(path1 is trial2)
        self.assertTrue(path1 is ensembles[4]['path_ensemble'].last_path)
        # Here, path2 and trial1 should be identical:
        self.assertTrue(path2 is trial1)
        self.assertTrue(path2 is ensembles[3]['path_ensemble'].last_path)
        # Copies should be identical except the
        # generated attribute and status.
        path1c.status = 'ACC'
        path2c.status = 'ACC'
        self.assertTrue(compare_path_skip_generated(path1, path1c))
        self.assertFalse(path1.generated[0] == path1c.generated[0])
        self.assertTrue(compare_path_skip_generated(path2, path2c))
        self.assertFalse(path2.generated[0] == path2c.generated[0])

    def test_swap_internal_wf(self):
        """Test swapping of paths with the wf move."""
        settings, ensembles = create_ensembles_and_paths()
        settings['tis']['high_accept'] = True

        with tempfile.TemporaryDirectory() as tempdir:
            # Test retis_swap_zero() that will perform sh in both 000 and 001,
            # but we define 001 to be a WF HA ensemble, hence high_acc_swap
            # has to be called.
            settings['ensemble'][1]['tis']['shooting_move'] = 'wf'
            settings['ensemble'][1]['tis']['high_accept'] = True
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            ensembles[1]['engine'].time = 10
            accept, (_, _), status = retis_swap_zero(
                ensembles, settings=settings, cycle=0)
            self.assertFalse(accept)
            self.assertEqual(status, 'HAS')
            accept, (trial1, trial2), status = retis_swap_zero(
                ensembles, settings=settings, cycle=0)
            self.assertTrue(accept)
            self.assertEqual(status, 'ACC')
            self.assertEqual(trial1.weight, 1)
            self.assertEqual(trial2.weight, 5)

            # We test retis_swap() with WF HA in both 003 and 004 ensembles
            for i in [3, 4]:
                settings['ensemble'][i]['tis']['shooting_move'] = 'wf'
                settings['ensemble'][i]['tis']['high_accept'] = True
                make_dirs(os.path.join(tempdir, f'00{i}'))
                settings['ensemble'][i]['simulation']['exe_path'] = tempdir
            settings['ensemble'][4]['tis']['interface_cap'] = 5.0
            accept, (trial1, trial2), status = retis_swap(
                ensembles, idx=3, settings=settings, cycle=0)
            self.assertTrue(accept)
            self.assertEqual(status, 'ACC')
            self.assertEqual(trial1.weight, 120)
            self.assertEqual(trial2.weight, 37)

    def test_swap_internal_repptis_SWD(self):
        """Test swapping of paths with repptis
        1) Try [0^+-'] <--> [1^+-]:
        1) Try    LML  <-->   LMR
        This move should fail due to incompatible swap directions (SWD)
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis')

        path1 = ensembles[1]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[2]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '001'))
            make_dirs(os.path.join(tempdir, '002'))
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            settings['ensemble'][2]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=1, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'SWD')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertTrue(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertTrue(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_0minus_fail(self):
        """Test swapping [0-] and [0+-'] in repptis.
        A RML path in [0+-'] cannot be swapped with [0-] (at least not for
        the current implementation.

        """
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         number=1)
        ensembles[0]['rgen'].seed = 5
        path1 = ensembles[0]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[1]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=0, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'SWD')
        # Check that the return trial paths have changed.
        self.assertTrue(compare_path_skip_generated(path1, trial2))
        self.assertTrue(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(path2.generated == trial1.generated)
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_SWH(self):
        """Test swapping of paths with repptis
        2) Try [1^+] with [2^+]
        2) Try   LMR <--> LML
        Depending on the seed for ens 0, this move is accepted or not.
        Yes, ensemble 0, as the prop_dir draw depends on that (there are
        issues with seed coupling in the random generator. To avoid coupling,
        both ensembles in the swap move drew from the same generator.)

        2A) Good prop direction, but first path propagation exceeds maxlen.
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis')
        ensembles[0]['rgen'].seed = 5
        path1 = ensembles[2]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[3]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '002'))
            make_dirs(os.path.join(tempdir, '003'))
            settings['ensemble'][2]['simulation']['exe_path'] = tempdir
            settings['ensemble'][3]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=2, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'SWH')
        # Check that the return trial path [i+-] has changed, while the
        # trial1 path is identical to the last accepted path of[i+-].
        self.assertTrue(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_ACC0(self):
        """Test swapping of paths with repptis
        2) Try [1^+] with [2^+]
        2) Try   LMR <--> LML
        Depending on the seed for ens 0, this move is accepted or not.
        Yes, ensemble 0, as the prop_dir draw depends on that (there are
        issues with seed coupling in the random generator. To avoid coupling,
        both ensembles in the swap move drew from the same generator.)

        2B) Good prop direction, and it works
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         total_eclipse=10000)
        ensembles[0]['rgen'].seed = 4
        path1 = ensembles[2]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[3]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '002'))
            make_dirs(os.path.join(tempdir, '003'))
            settings['ensemble'][2]['simulation']['exe_path'] = tempdir
            settings['ensemble'][3]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=2, settings=settings, cycle=0)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Check that the return trial path [i+-] has changed, while the
        # trial1 path is identical to the last accepted path of[i+-].
        self.assertFalse(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_ACC1(self):
        """Test swapping of paths with repptis
        3) Swap [3^+] <--> [4^+]
        3) Swap   RMR <--> LML: always acceptable propdirs
        3) We change total_eclipse to a very large number to force
           This move to be accepted.
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         total_eclipse=10000)
        path1 = ensembles[4]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[5]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '004'))
            make_dirs(os.path.join(tempdir, '005'))
            settings['ensemble'][4]['simulation']['exe_path'] = tempdir
            settings['ensemble'][5]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=4, settings=settings, cycle=0)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertFalse(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_ACC2(self):
        """Test swapping of paths with repptis
        3) Swap [3^+] <--> [4^+]
        3) Swap   RMR <--> LML: always acceptable propdirs
        3) We change total_eclipse to a very large number to force
           This move to be accepted.
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         total_eclipse=10000)
        ensembles[0]['rgen'].seed = 1
        path1 = ensembles[4]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[5]['path_ensemble'].last_path
        path2c = path2.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '004'))
            make_dirs(os.path.join(tempdir, '005'))
            settings['ensemble'][4]['simulation']['exe_path'] = tempdir
            settings['ensemble'][5]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=4, settings=settings, cycle=0)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertFalse(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

    def test_swap_internal_repptis_FTX(self):
        """Test swapping of paths with repptis
        4) Swap [3^+] <--> [4^+]
        4) Swap   RMR <--> LML: always acceptable propdirs
        4) We total_eclipse to a very large number to force
        4A) but we decrease maxlen to force BTX/FTX in first extension
        4B) but we decrease maxlen to force BTX/FTX in second extension

        4A) too long in first extension
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         total_eclipse=10000)
        path1 = ensembles[4]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[5]['path_ensemble'].last_path
        path2c = path2.copy()
        # decrease maxlen in ensemble 4
        settings['ensemble'][4]['tis']['maxlength'] = 200
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '004'))
            make_dirs(os.path.join(tempdir, '005'))
            settings['ensemble'][4]['simulation']['exe_path'] = tempdir
            settings['ensemble'][5]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=4, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'SWH')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertTrue(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that trial1 reached the maximum length:
        self.assertEqual(len(trial1.phasepoints), 200)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

        # 4B) too long in second extension
        settings, ensembles = create_ensembles_and_paths(task='repptis',
                                                         total_eclipse=10000)
        path1 = ensembles[4]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[5]['path_ensemble'].last_path
        path2c = path2.copy()
        # decrease maxlen in ensemble 5
        settings['ensemble'][5]['tis']['maxlength'] = 200
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '004'))
            make_dirs(os.path.join(tempdir, '005'))
            settings['ensemble'][4]['simulation']['exe_path'] = tempdir
            settings['ensemble'][5]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap_wrapper(
                ensembles, idx=4, settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'FTX')
        # Check that the return trial paths are identical
        # to the accepted paths, with the exception of the move:
        self.assertFalse(compare_path_skip_generated(path1, trial2))
        self.assertFalse(path1.generated == trial2.generated)
        self.assertFalse(compare_path_skip_generated(path2, trial1))
        self.assertFalse(path2.generated == trial1.generated)
        # Check that trial2 reached the maximum length:
        self.assertEqual(len(trial2.phasepoints), 200)
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)

        # Testing for BTS and FTS makes no sense for REPPTIS swap, as it is
        # impossible by definition.

    def test_high_acc_ss_wt_swap(self):
        """Test weights and high_acc_swap method for the ss and wt move."""
        settings, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=3)
        interfaces = [-1., 0., 1., 2., 10]
        # 1) Try [0^+] with [1^+]:
        # This move should be reject, we here check that
        # the currently accepted paths are not modified.
        path1 = ensembles[2]['path_ensemble'].last_path.copy()
        path2 = ensembles[4]['path_ensemble'].last_path.copy()
        intf0 = [interfaces[0], interfaces[1], interfaces[-1]]
        intf1 = [interfaces[0], interfaces[2], interfaces[-1]]

        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['ss', 'ss'])
        self.assertFalse(success)
        self.assertEqual(status, 'HAS')

        path1 = ensembles[3]['path_ensemble'].last_path.copy()
        path2 = ensembles[4]['path_ensemble'].last_path.copy()
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['ss', 'ss'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        # Now let's check with ss only in path1
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['ss', 'wt'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        path1 = ensembles[4]['path_ensemble'].last_path.copy()
        path2 = ensembles[4]['path_ensemble'].last_path.copy()
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['ss', 'wt'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        # Now let's check with ss only in path2
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['wt', 'ss'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        path1 = ensembles[4]['path_ensemble'].last_path.copy()
        path2 = ensembles[4]['path_ensemble'].last_path.copy()
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['wt', 'ss'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        path1 = ensembles[3]['path_ensemble'].last_path.copy()
        success, status = high_acc_swap([path1, path2], rgen,
                                        intf0,
                                        intf1,
                                        ['wt', 'ss'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

        settings['tis']['high_accept'] = True
        fill_up_tis_and_retis_settings(settings)
        accept, (_, _), status = retis_swap(ensembles,
                                            idx=3,
                                            settings=settings,
                                            cycle=1)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')

        # Check that we can use the ss weights:
        ensembles[3]['path_ensemble'].last_path.set_move('ss')
        ensembles[4]['path_ensemble'].last_path.set_move('ss')
        settings['ensemble'][4]['tis']['shooting_move'] = 'ss'
        accept, (_, _), status = retis_swap(ensembles, idx=3,
                                            settings=settings,
                                            cycle=1)
        self.assertEqual(ensembles[4]['path_ensemble'].last_path.weight, 2)

    def test_high_acc_wf_swap(self):
        """Test weights and high_acc_swap method for the wf move."""
        _, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=3)
        interfaces = [-1., 0., 1., 2., 10]
        # 1) Try [0^+] with [1^+]:
        path1 = ensembles[2]['path_ensemble'].last_path.copy()
        path2 = ensembles[4]['path_ensemble'].last_path.copy()
        intf1 = [interfaces[0], interfaces[1], interfaces[-1]]
        intf2 = [interfaces[0], interfaces[2], interfaces[-1]]

        c1_old = compute_weight(path1, intf1, 'wf')
        c2_old = compute_weight(path2, intf2, 'wf')
        c1_new = compute_weight(path1, intf2, 'wf')
        c2_new = compute_weight(path2, intf1, 'wf')
        self.assertEqual([c1_old, c2_old, c1_new, c2_new], [54, 120, 0, 120])
        success, status = high_acc_swap([path1, path2], rgen, intf1,
                                        intf2, ['wf', 'wf'])
        self.assertFalse(success)
        self.assertEqual(status, 'HAS')

        success, status = high_acc_swap([path1, path2], rgen, intf2,
                                        intf1, ['wf', 'wf'])
        self.assertTrue(success)
        self.assertEqual(status, 'ACC')

    def test_nullmove_internal(self):
        """Test the null move."""
        _, ensembles = create_ensembles_and_paths()
        for ens in ensembles:
            path0 = ens['path_ensemble'].last_path
            before = ens['path_ensemble'].last_path.copy()
            accept, trial, status = null_move(ens, 1)
            self.assertTrue(accept)
            self.assertTrue(path0 is trial)
            self.assertTrue(status == 'ACC')
            self.assertTrue(path0.generated[0] == '00')
            after = ens['path_ensemble'].last_path
            self.assertTrue(compare_path_skip_generated(before, after))

    def test_swap_zero_internal_repptis(self):
        """Test the repptis swap zero move."""
        settings, ensembles = create_ensembles_and_paths()
        settings['simulation']['task'] = 'repptis'

        path1 = ensembles[0]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[1]['path_ensemble'].last_path
        path2c = path2.copy()
        ensembles[1]['engine'].time = 6
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = repptis_swap(ensembles,
                                                            idx=0,
                                                            settings=settings,
                                                            cycle=0)
        # This should be accepted:
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)
        # Last point in trial 1 is second in path 2:
        self.assertEqual(trial1.phasepoints[-1], path2.phasepoints[1])
        # Second last point in trial 1 is first in path 2:
        self.assertEqual(trial1.phasepoints[-2], path2.phasepoints[0])
        # First point in trial 2 is second last in path 1:
        self.assertEqual(trial2.phasepoints[0], path1.phasepoints[-2])
        # Second point in trial 2 is last point in path 1:
        self.assertEqual(trial2.phasepoints[1], path1.phasepoints[-1])

    def test_swap_zero_internal(self):
        """Test the retis swap zero move."""
        settings, ensembles = create_ensembles_and_paths()

        path1 = ensembles[0]['path_ensemble'].last_path
        path1c = path1.copy()
        path2 = ensembles[1]['path_ensemble'].last_path
        path2c = path2.copy()
        ensembles[1]['engine'].time = 6
        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            accept, (trial1, trial2), status = retis_swap(ensembles,
                                                          idx=0,
                                                          settings=settings,
                                                          cycle=0)
        # This should be accepted:
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        # Check that paths path1 and path2 did not change:
        self.assertEqual(path1, path1c)
        self.assertEqual(path2, path2c)
        # Last point in trial 1 is second in path 2:
        self.assertEqual(trial1.phasepoints[-1], path2.phasepoints[1])
        # Second last point in trial 1 is first in path 2:
        self.assertEqual(trial1.phasepoints[-2], path2.phasepoints[0])
        # First point in trial 2 is second last in path 1:
        self.assertEqual(trial2.phasepoints[0], path1.phasepoints[-2])
        # Second point in trial 2 is last point in path 1:
        self.assertEqual(trial2.phasepoints[1], path1.phasepoints[-1])

    def test_swap_zero_internal_0L(self):
        ens0 = PathEnsemble(ensemble_number=0, interfaces=(0, 2, 2))
        ens1 = PathEnsemble(ensemble_number=1, interfaces=(2, 2.5, 3))
        path0 = make_internal_path((0, 2.1), (100, 2.2), (50, -1),
                                   ens0.interfaces[1])
        path1 = make_internal_path((0, 1.9), (100, 1.8), (50, 2.5),
                                   ens1.interfaces[1])

        ens0.add_path_data(path0, status='ACC')
        ens1.add_path_data(path1, status='ACC')
        ens0.last_path = path0
        ens1.last_path = path1

        order_f = MockOrder()
        engine = MockEngine(10, turn_around=15)
        engine.delta_v *= -1
        ensemble0 = {'path_ensemble': ens0, 'engine': engine,
                     'order_function': order_f, 'interfaces': [0, 2, 2]}
        ensemble1 = {'path_ensemble': ens1, 'engine': engine,
                     'order_function': order_f, 'interfaces': [2, 2.5, 3]}
        with tempfile.TemporaryDirectory() as tpdr:
            make_dirs(os.path.join(tpdr, '000'))
            make_dirs(os.path.join(tpdr, '001'))
            out = retis_swap_zero(
                ensembles=[ensemble0, ensemble1],
                settings={'ensemble': [{'tis': {'maxlength': 1000},
                                        'simulation': {'exe_path': tpdr}}]*2},
                cycle=1)
        self.assertFalse(out[0])
        self.assertEqual(out[1][0].status, "0-L")
        self.assertEqual(out[1][1].status, "0-L")
        self.assertEqual(out[1][0].check_interfaces(ens0.interfaces)[:2],
                         ('L', 'R'))

    def test_swap_zero_internal_0L_acc(self):
        ens0 = PathEnsemble(ensemble_number=0, interfaces=(0, 2, 2))
        ens1 = PathEnsemble(ensemble_number=1, interfaces=(2, 2.5, 3))
        path0 = make_internal_path((0, 2.1), (100, 2.2), (50, -1),
                                   ens0.interfaces[1])
        path1 = make_internal_path((0, 1.9), (100, 1.8), (50, 2.5),
                                   ens1.interfaces[1])

        ens0.add_path_data(path0, status='ACC')
        ens1.add_path_data(path1, status='ACC')
        ens0.last_path = path0
        ens1.last_path = path1

        order_f = MockOrder()
        engine = MockEngine(5, turn_around=100)
        ensemble0 = {'path_ensemble': ens0, 'engine': engine,
                     'order_function': order_f, 'interfaces': [0, 2, 2]}
        ensemble1 = {'path_ensemble': ens1, 'engine': engine,
                     'order_function': order_f, 'interfaces': [2, 2.5, 3]}
        with tempfile.TemporaryDirectory() as tpdr:
            make_dirs(os.path.join(tpdr, '000'))
            make_dirs(os.path.join(tpdr, '001'))
            out = retis_swap_zero(
                ensembles=[ensemble0, ensemble1],
                settings={'ensemble': [{'tis': {'maxlength': 1000},
                                        'simulation': {'exe_path': tpdr}}]*2,
                          'tis': {'high_accept': False}},
                cycle=1)
        self.assertTrue(out[0])
        self.assertEqual(out[1][0].status, "ACC")
        self.assertEqual(out[1][1].status, "ACC")
        self.assertEqual(out[1][0].check_interfaces(ens0.interfaces)[:2],
                         ('R', 'R'))

    def test_swap_zero_internal_ftx(self):
        """Test the swap zero when we force a FTX"""
        settings, ensembles = create_ensembles_and_paths()
        with tempfile.TemporaryDirectory() as tempdir:
            ensembles[0]['engine'] = MockEngine(timestep=1.0, turn_around=500)
            settings['ensemble'][0]['tis']['maxlength'] = 100
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            accept, _, status = retis_swap(ensembles, idx=0,
                                           settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'FTX')

    def test_swap_zero_internal_btx(self):
        """Test the swap zero when we force a BTX"""
        settings, ensembles = create_ensembles_and_paths()
        for i_ens, ens in enumerate(ensembles):
            ens['engine'] = MockEngine(timestep=1.0, turn_around=500)
            settings['ensemble'][i_ens]['tis']['maxlength'] = 3
            with tempfile.TemporaryDirectory() as tempdir:
                settings['ensemble'][0]['simulation']['exe_path'] = tempdir
                settings['ensemble'][1]['simulation']['exe_path'] = tempdir
                make_dirs(os.path.join(tempdir, '000'))
                make_dirs(os.path.join(tempdir, '001'))
                accept, _, status = retis_swap(ensembles, idx=0,
                                               settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'BTX')

    def test_swap_zero_internal_bts(self):
        """Test the swap zero when we force a BTS"""
        settings, ensembles = create_ensembles_and_paths()
        for ens in ensembles:
            ens['engine'] = MockEngine(timestep=200.0)
        # We set up for BTS by making a faulty initial path:
        path = make_internal_path((0, -0.9), (100, -1.2), (50, -0.2), None)
        ensembles[1]['path_ensemble'].add_path_data(path, 'ACC')
        accept, _, status = retis_swap(ensembles, idx=0,
                                       settings=settings, cycle=1)
        self.assertFalse(accept)
        self.assertEqual(status, 'BTS')

    def test_swap_zero_internal_non_accept(self):
        """Test the swap zero when we force a FTS"""
        settings, ensembles = create_ensembles_and_paths()
        # We set up for FTS by making a faulty initial path:
        with tempfile.TemporaryDirectory() as tempdir:
            settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings['ensemble'][1]['simulation']['exe_path'] = tempdir
            make_dirs(os.path.join(tempdir, '000'))
            make_dirs(os.path.join(tempdir, '001'))
            path = make_internal_path((0, -0.9), (100, -1.2), (50, -5), None)
            ensembles[0]['path_ensemble'].add_path_data(path, 'ACC')
            accept, _, status = retis_swap(ensembles, idx=0,
                                           settings=settings, cycle=0)
        self.assertFalse(accept)
        self.assertEqual(status, 'BTS')

    def test_retis_moves(self):
        """Test the retis moves function."""
        settings, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=0)
        settings['retis']['swapsimul'] = False
        settings['retis']['nullmoves'] = True
        path1 = ensembles[3]['path_ensemble'].last_path
        path2 = ensembles[4]['path_ensemble'].last_path
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        # We should have done swapping for [2^+] and [3^+] and nullmoves
        # for the rest:
        for resi in results:
            idx = resi['ensemble_number']
            if idx not in (3, 4):
                self.assertEqual(resi['mc-move'], 'nullmove')
            else:
                self.assertEqual(resi['mc-move'], 'swap')
                if idx == 3:
                    self.assertEqual(resi['swap-with'], 4)
                    self.assertTrue(path2 is resi['trial'])
                elif idx == 4:
                    self.assertEqual(resi['swap-with'], 3)
                    self.assertTrue(path1 is resi['trial'])
            self.assertEqual(resi['status'], 'ACC')
            self.assertTrue(resi['accept'])

    def test_repptis_moves(self):
        """Test the repptis moves function, NO swapsimul.
        We have 6 ensembles, with the following acceptable paths:
            [0^-]---RMR
            [0^+-']-LML
            [1^+-]--LMR
            [2^+-]--LML
            [3^+-]--RMR
            [4^+-]--LML
        Ensembles 2^+- and 3^+- are swapped, the others do nullmoves
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis')
        rgen = MockRandomGenerator(seed=0)
        settings['retis']['swapsimul'] = False
        settings['retis']['nullmoves'] = True
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        # for the rest:
        statuses = ['SWD', 'SWD', 'ACC', 'ACC', 'ACC', 'ACC']
        mc_moves = ['swap', 'swap', 'nullmove', 'nullmove', 'nullmove',
                    'nullmove']
        ens_names = ['[2^+]', '[3^+]', '[0^-]', '[0^+]', '[1^+]', '[4^+]']
        for i, resi in enumerate(results):
            idx = resi['ensemble_number']
            if idx == 3:
                self.assertEqual(resi['swap-with'], 4)
            elif idx == 4:
                self.assertEqual(resi['swap-with'], 3)
            self.assertTrue(resi['mc-move'] in ['nullmove', 'swap'])
            self.assertEqual(ensembles[idx]['path_ensemble'].ensemble_name,
                             ens_names[i])
            self.assertEqual((idx, resi['status']), (idx, statuses[i]))
            self.assertEqual(resi['mc-move'], mc_moves[i])

    def test_repptis_moves_simul(self):
        """Test the repptis moves function.
        We have 6 ensembles, with the following acceptable paths:
            [0^-]---RMR
            [0^+-']-LML
            [1^+-]--LMR
            [2^+-]--LML
            [3^+-]--RMR
            [4^+-]--LML
        """
        settings, ensembles = create_ensembles_and_paths(task='repptis')
        rgen = MockRandomGenerator(seed=0)
        settings['retis']['swapsimul'] = True
        settings['retis']['nullmoves'] = True
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        # for the rest:
        statuses = ['SWD', 'SWD', 'SWD', 'SWD', 'ACC', 'ACC']
        mc_moves = ['swap', 'swap', 'swap', 'swap', 'nullmove', 'nullmove']
        ens_names = ['[0^+]', '[1^+]', '[2^+]', '[3^+]', '[4^+]', '[0^-]']
        for i, resi in enumerate(results):
            idx = resi['ensemble_number']
            if idx == 3:
                self.assertEqual(resi['swap-with'], 4)
            elif idx == 4:
                self.assertEqual(resi['swap-with'], 3)
            self.assertTrue(resi['mc-move'] in ['nullmove', 'swap'])
            self.assertEqual(ensembles[idx]['path_ensemble'].ensemble_name,
                             ens_names[i])
            self.assertEqual((idx, resi['status']), (idx, statuses[i]))
            self.assertEqual(resi['mc-move'], mc_moves[i])

    def test_retis_moves_simul(self):
        """Test the retis moves function with swaps."""
        settings, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=0)
        settings['retis']['swapsimul'] = True
        settings['retis']['nullmoves'] = True
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        # We expect a nullmove for the first and swapping for the rest:
        moves = ('nullmove', 'swap', 'swap', 'swap', 'swap')
        for resi in results:
            self.assertEqual(resi['mc-move'],
                             moves[resi['ensemble_number']])
        # Try with an even number of ensembles. This should trigger
        # the ``if len(ensembles) % 2`` for a particular scheme, we
        # enforce this scheme by resetting the seed:
        ensembles = ensembles[:-1]
        ensembles[0]['path_ensemble'].last_path.set_move('ld')
        rgen = MockRandomGenerator(seed=0)
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        moves = ('nullmove', 'swap', 'swap', 'nullmove')
        for resi in results:
            self.assertEqual(resi['mc-move'],
                             moves[resi['ensemble_number']])
        self.assertEqual(ensembles[0]['path_ensemble'].last_path.get_move(),
                         'ld')
        # Finally, try with just 2 ensembles:
        ensembles = ensembles[0:2]
        results = retis_moves(ensembles, rgen, settings, cycle=1)
        for resi in results:
            self.assertEqual(resi['mc-move'], 'swap')

    def test_relative_shoots(self):
        """Test the relative shoots selection."""
        _, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=0)
        relative = [0.1, 0.1, 0.1, 0.1, 0.6]
        idx, ensemble = relative_shoots_select(ensembles, rgen, relative)
        self.assertEqual(idx, 4)
        self.assertEqual(ensemble, ensembles[idx])
        relative = [1.0, 0.0, 0.0, 0.0, 0.0]
        idx, ensemble = relative_shoots_select(ensembles, rgen, relative)
        self.assertEqual(idx, 0)
        self.assertEqual(ensemble, ensembles[idx])
        relative = [0.0, 0.0, 0.0, 0.0, 0.0]
        with self.assertRaises(ValueError):
            relative_shoots_select(ensembles, rgen, relative)

    def test_make_retis_step(self):
        """Test that we can do the RETIS steps."""
        settings, ensembles = create_ensembles_and_paths()
        rgen = MockRandomGenerator(seed=0)
        for setting in settings['ensemble']:
            setting['tis']['freq'] = 1.0
        # Check that we can do RETIS:
        settings['retis']['swapfreq'] = 1.0
        settings['retis']['swapsimul'] = True
        settings['retis']['nullmoves'] = True
        ensembles[1]['path_ensemble'].last_path.set_move('ld')
        results = make_retis_step(ensembles, rgen, settings, cycle=1)
        self.assertTrue(ensembles[1]['path_ensemble'].last_path.get_move() ==
                        'ld')
        moves = ('swap', 'swap', 'swap', 'swap', 'nullmove')
        for resi in results:
            self.assertEqual(resi['mc-move'],
                             moves[resi['ensemble_number']])
        # Check that we can select TIS moves:
        settings['retis']['swapfreq'] = 0.0
        results = make_retis_step(ensembles, rgen, settings, cycle=1)

        moves = ('tr', 'ld', 'tr', 'tr', 'tr')
        for resi in results:
            self.assertEqual(
                resi['mc-move'], moves[resi['ensemble_number']])

        # Check that we can do relative shoots:
        settings['retis']['relative_shoots'] = [0.1, 0.1, 0.1, 0.1, 0.6]

        with tempfile.TemporaryDirectory() as tempdir:
            settings['engine']['exe_path'] = tempdir
            settings['simulation']['exe_path'] = tempdir
            for ens in settings['ensemble']:
                ens['engine']['exe_path'] = tempdir
                ens['simulation']['exe_path'] = tempdir
            for fd in {'000', '001', '002', '003', '004'}:
                make_dirs(os.path.join(tempdir, fd))
            results2 = make_retis_step(ensembles, rgen, settings, cycle=0)
            moves = ('nullmove', 'nullmove', 'nullmove', 'nullmove', 'tr')
            for resi in results2:
                self.assertEqual(
                    resi['mc-move'], moves[resi['ensemble_number']])

        # Check that we can do priority shooting:
        settings['retis']['relative_shoots'] = None
        settings['simulation']['priority_shooting'] = True
        results2 = make_retis_step(ensembles, rgen, settings, cycle=2)
        ensembles[1]['path_ensemble'].nstats['npath'] = 1
        moves = (None, 'ld', None, None, None)
        for i, resi in enumerate(results2):
            if i != 1:
                self.assertEqual(resi, moves[i])
            else:
                self.assertEqual(resi['mc-move'],
                                 moves[resi['ensemble_number']])

    def test_make_repptis_step(self):
        """Test that we can do REPPTIS steps."""
        settings, ensembles = create_ensembles_and_paths(task='repptis')
        rgen = MockRandomGenerator(seed=1)
        for setting in settings['ensemble']:
            setting['tis']['freq'] = 1.0
        # Check that we can do REPPTIS:
        settings['retis']['swapfreq'] = 1.0
        settings['retis']['swapsimul'] = True
        settings['retis']['nullmoves'] = True
        ensembles[1]['path_ensemble'].last_path.set_move('ld')
        results = make_retis_step(ensembles, rgen, settings, cycle=1)
        self.assertTrue(ensembles[1]['path_ensemble'].last_path.get_move() ==
                        'ld')
        moves = ('nullmove', 'swap', 'swap', 'swap', 'swap', 'nullmove')
        for resi in results:
            self.assertEqual(resi['mc-move'],
                             moves[resi['ensemble_number']])

        # Check that we can select TIS moves:
        settings['retis']['swapfreq'] = 0.0
        results = make_retis_step(ensembles, rgen, settings, cycle=1)

        moves = ('tr', 'ld', 'tr', 'tr', 'tr', 'tr')
        for resi in results:
            self.assertEqual(
                resi['mc-move'], moves[resi['ensemble_number']])

        # Check that we can do relative shoots:
        settings['retis']['relative_shoots'] = [0.1, 0.1, 0.1, 0.1, 0.6, 0.0]

        with tempfile.TemporaryDirectory() as tempdir:
            settings['engine']['exe_path'] = tempdir
            settings['simulation']['exe_path'] = tempdir
            for ens in settings['ensemble']:
                ens['engine']['exe_path'] = tempdir
                ens['simulation']['exe_path'] = tempdir
            for fd in {'000', '001', '002', '003', '004', '005'}:
                make_dirs(os.path.join(tempdir, fd))
            results2 = make_retis_step(ensembles, rgen, settings, cycle=0)
            moves = ('nullmove', 'nullmove', 'nullmove', 'nullmove', 'tr',
                     'nullmove')
            for resi in results2:
                self.assertEqual(
                    resi['mc-move'], moves[resi['ensemble_number']])

        # Check that we can do priority shooting:
        settings['retis']['relative_shoots'] = None
        settings['simulation']['priority_shooting'] = True
        results2 = make_retis_step(ensembles, rgen, settings, cycle=2)
        ensembles[1]['path_ensemble'].nstats['npath'] = 1
        moves = (None, 'ld', None, None, None, None)
        for i, resi in enumerate(results2):
            if i != 1:
                self.assertEqual(resi, moves[i])
            else:
                self.assertEqual(resi['mc-move'],
                                 moves[resi['ensemble_number']])

    def test_partial_exit_retis(self):
        """Test that we can quit from the RETIS steps."""
        sim, settings = prepare_test_simulation()
        with tempfile.TemporaryDirectory() as tempdir:
            # Run a healthy simulation
            sim.settings['simulation']['exe_path'] = tempdir
            sim.settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            with patch('sys.stdout', new=StringIO()):
                for _ in initiate_path_simulation(sim, settings):
                    logging.debug('Running initialisation')
            make_dirs(os.path.join(tempdir, '002'))
            for _ in sim.run():
                pass
            self.assertEqual(
                sim.ensembles[0]['path_ensemble'].paths[-1:][0]['cycle'], 10)

        del sim
        sim, settings = prepare_test_simulation()
        with tempfile.TemporaryDirectory() as tempdir:
            sim.settings['simulation']['exe_path'] = tempdir
            sim.settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            with patch('sys.stdout', new=StringIO()):
                for _ in initiate_path_simulation(sim, settings):
                    logging.debug('Running initialisation')
            make_dirs(os.path.join(tempdir, '002'))

            self.assertEqual(
                sim.ensembles[0]['path_ensemble'].paths[-1:][0]['cycle'], 0)
            sim.step()
            self.assertEqual(
                sim.ensembles[0]['path_ensemble'].paths[-1:][0]['cycle'], 1)

            # Insert Covid19 and check its presence
            os.mknod(os.path.join(tempdir, "EXIT"))
            with patch('sys.stdout', new=StringIO()):
                self.assertTrue(soft_partial_exit(tempdir))

            # Run a zombie simulation
            with patch('sys.stdout', new=StringIO()):
                for _ in sim.run():
                    logging.info('Try to walk')
            self.assertEqual(
                sim.ensembles[0]['path_ensemble'].paths[-1:][0]['cycle'], 2)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting path-trajectory data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PathExtFormatter (:py:class:`.PathExtFormatter`)
    A class for formatting external trajectories. Since this format
    is intended for paths created by an external software it will
    simply contain locations for the files where the snapshots are
    stored.

PathIntFormatter (:py:class:`.PathIntFormatter`)
    A class for formatting internal trajectories. The format used is
    relatively simple and does not contain information about atoms,
    just the coordinates. The output for n-dimensional system contains
    2n columns (positions and velocities). This formatter is intended
    for use with objects like :py:class:`.Path`.

PathExtFile (:py:class:`.PathExtFile`)
    A class for writing path data for external paths.

PathIntFile (:py:class:`.PathIntFile`)
    A class for writing path data for internal paths.

"""
import logging
import os
import numpy as np
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO, read_some_lines
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'PathExtFormatter',
    'PathIntFormatter',
    'PathExtFile',
    'PathIntFile',
]


class PathIntFormatter(OutputFormatter):
    """A class for formatting internal trajectories."""

    _FMT = '{:15.9f}'

    def __init__(self):
        """Initialise the PathIntFormatter formatter."""
        super().__init__('PathIntFormatter', header=None)
        self.print_header = False
        self.fmt = None

    def format(self, step, data):
        """Format path data for internal paths.

        Parameters
        ----------
        step : integer
            The current simulation step.
        data : list
            Here, ``data[0]`` is assumed to be an object like
            :py:class:`.Path` and ``data[1]`` a string containing the
            status of this path.

        Yields
        ------
        out : string
            The trajectory (positions & velocities).

        """
        path, status = data[0], data[1]
        if not path:  # E.g. when null-moves are False.
            return
        yield f'# Cycle: {step}, status: {status}'
        for i, phasepoint in enumerate(path.phasepoints):
            yield f'Snapshot: {i}'
            pos = phasepoint.particles.get_pos()
            vel = phasepoint.particles.get_vel()
            for posj, velj in zip(pos, vel):
                if self.fmt is None:
                    self.fmt = ' '.join([self._FMT] * len(posj))
                yield ' '.join([self.fmt.format(*posj),
                                self.fmt.format(*velj)])

    @staticmethod
    def parse(line):
        """Parse positions and velocities from a single line.

        Parameters
        ----------
        line : string
            The line to parse.

        Returns
        -------
        pos : list of floats
            The positions read from the file.
        vel : list of floats
            The velocities read from the file.

        """
        raw = [float(i) for i in line.split()]
        dim = len(raw) // 2
        if not 1 <= dim <= 3:
            raise ValueError(f'Malformed trajectory data: dim = {dim}?')
        pos = raw[:dim]
        vel = raw[dim:2*dim]
        return pos, vel

    def read_snapshots(self, data):
        """Read snapshots from data.

        Parameters
        ----------
        data : list of strings
            The data we are reading into snapshots.

        Yields
        ------
        out : dict
            A dict representing the snapshot with positions and
            velocities.

        """
        pos, vel = [], []
        for line in data:
            if line.startswith('Snapshot'):
                # A new snapshot, store the old one:
                if pos:
                    yield {'pos': np.array(pos), 'vel': np.array(vel)}
                pos, vel = [], []
            else:
                new_pos, new_vel = self.parse(line)
                pos.append(new_pos)
                vel.append(new_vel)
        if pos:
            yield {'pos': np.array(pos), 'vel': np.array(vel)}

    def load(self, filename):
        """Load trajectories from a path file.

        Parameters
        ----------
        filename : string
            The path/file name of the file we want to open.

        Yields
        ------
        data : dict
            The trajectory & comment read from the path file.

        """
        for trajectory in read_some_lines(filename, line_parser=None):
            # These trajectories are just raw text,
            # convert to snapshots:
            traj = list(self.read_snapshots(trajectory['data']))
            data = {'comment': trajectory['comment'],
                    'data': traj}
            yield data


class PathExtFormatter(OutputFormatter):
    """A class for formatting external trajectories.

    The external trajectories as stored as files and this path
    formatter includes the location of these files.

    Attributes
    ----------
    FMT : string
        The string to use for the formatting.

    """

    FMT = '{:>10}  {:>20s}  {:>10}  {:>5}'

    def __init__(self):
        """Initialise the PathExtFormatter formatter."""
        header = {'labels': ['Step', 'Filename', 'index', 'vel'],
                  'width': [10, 20, 10, 5], 'spacing': 2}

        super().__init__('PathExtFormatter', header=header)
        self.print_header = False

    def format(self, step, data):
        """Format path data for external paths.

        Parameters
        ----------
        step : integer
            The current simulation step.
        data : list
            Here, ``data[0]`` is assumed to be an object like
            :py:class:`.Path`` and ``data[1]`` a string containing the
            status of this path.

        Yields
        ------
        out : string
            The trajectory as references to files.

        """
        path, status = data[0], data[1]
        if not path:  # E.g. when null-moves are False.
            return
        yield f'# Cycle: {step}, status: {status}'
        yield self.header
        for i, phasepoint in enumerate(path.phasepoints):
            filename, idx = phasepoint.particles.get_pos()
            filename_short = os.path.basename(filename)
            if idx is None:
                idx = 0
            vel = -1 if phasepoint.particles.get_vel() else 1
            yield self.FMT.format(i, filename_short, idx, vel)

    @staticmethod
    def parse(line):
        """Parse the line data by splitting text on spaces.

        Parameters
        ----------
        line : string
            The line to parse.

        Returns
        -------
        out : list
            The columns of data.

        """
        return list(line.split())


class PathExtFile(FileIO):
    """A class for writing path data."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the path writer with correct format for external paths."""
        super().__init__(filename, file_mode, PathExtFormatter(),
                         backup=backup)


class PathIntFile(FileIO):
    """A class for writing path data."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the path writer with correct format for internal paths."""
        super().__init__(filename, file_mode, PathIntFormatter(),
                         backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A custom engine for particle swarm optimization."""
import numpy as np
from pyretis.engines import MDEngine


class PSOEngine(MDEngine):
    """Perform particle swarm optimization."""

    def __init__(self, inertia, accp, accg):
        """Set up the engine.

        Parameters
        ----------
        intertia : float
            The intertia factor in the velocity equation
            of motion.
        accp : float
            The acceleration for the previous best term. "The congnitive term".
        accg : float
            The acceleration for the global best term. "The social term".

        """
        super().__init__(1, 'Particle Swarm Optimization')
        self.inertia = inertia
        self.accp = accp
        self.accg = accg
        self.pbest = None
        self.pbest_pot = None
        self.gbest = None
        self.gbest_pot = None

    def integration_step(self, system):
        """Perform one step for the PSO algorithm."""
        particles = system.particles
        if self.pbest is None:
            self.pbest = np.copy(particles.pos)
            self.pbest_pot = system.potential()
        if self.gbest is None:
            pot = system.potential()
            idx = np.argmin(pot)
            self.gbest = particles.pos[idx]
            self.gbest_pot = pot[idx]

        rnd1 = np.random.uniform()
        rnd2 = np.random.uniform()

        particles.vel = (self.inertia * particles.vel +
                         rnd1 * self.accp * (self.pbest - particles.pos) +
                         rnd2 * self.accg * (self.gbest - particles.pos))
        particles.pos += particles.vel
        particles.pos = system.box.pbc_wrap(particles.pos)

        pot = system.potential()

        # Update global?
        idx = np.argmin(pot)
        if pot[idx] < self.gbest_pot:
            self.gbest_pot = pot[idx]
            self.gbest = particles.pos[idx]
        # Update for individuals:
        idx = np.where(pot < self.pbest_pot)[0]
        self.pbest[idx] = np.copy(particles.pos[idx])
        self.pbest_pot[idx] = pot[idx]
        return self.gbest_pot, self.gbest



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the path simulation classes."""
from io import StringIO
import copy
import logging
import os
import tempfile
import unittest
import pathlib

from pyretis.inout.settings import (add_default_settings,
                                    fill_up_tis_and_retis_settings,)
from pyretis.setup.createsimulation import create_ensembles
from pyretis.simulation.path_simulation import (PathSimulation,
                                                SimulationTIS,
                                                SimulationRETIS,)
from unittest.mock import patch
from .help import TEST_SETTINGS

logging.disable(logging.CRITICAL)

HERE = os.path.abspath(os.path.dirname(__file__))

SETTINGS1 = {
    'simulation': {
        'task': 'tis',
        'steps': 10,
        'interfaces': [-0.9, -0.9, 1.0],
        'seed': 1,
        'rgen': 'rgen',
    },
    'system': {'units': 'lj', 'temperature': 0.1, 'dimensions': 1},
    'engine': {
        'class': 'Langevin',
        'gamma': 0.3,
        'high_friction': False,
        'seed': 321,
        'timestep': 0.2
    },
    'potential': [{'a': 1.0, 'b': 2.0, 'c': 0.0, 'class': 'DoubleWell'}],
    'orderparameter': {
        'class': 'Position',
        'dim': 'x',
        'index': 0,
        'name': 'Position',
        'periodic': False
    },
    'initial-path': {'method': 'kick'},
    'particles': {'position': {'input_file': 'initial.xyz'}},
    'retis': {
        'swapfreq': 0.5,
        'nullmoves': True,
        'swapsimul': True,
    },
    'tis': {
        'freq': 0.5,
        'maxlength': 2000,
        'aimless': True,
        'allowmaxlength': False,
        'sigma_v': -1,
        'seed': 1,
        'rgen': 'rgen',
        'zero_momentum': False,
        'rescale_energy': False,
    },
}


def create_variables(exe_path=None):
    """Create system, engine, order parameter etc."""
    settings = {key: copy.deepcopy(val) for key, val in TEST_SETTINGS.items()}
    settings['engine'] = {'class': 'Langevin', 'gamma': 0.3,
                          'high_friction': False,
                          'timestep': 0.2}
    settings['orderparameter'] = {'class': 'Position',
                                  'index': 0}

    if exe_path is not None:
        settings['simulation'] = {'exe_path': exe_path}
    add_default_settings(settings)
    settings['particles']['type'] = 'internal'
    settings['simulation']['interfaces'] = [-1.0, 0.0, 1.0, 2.0]
    settings['simulation']['task'] = 'tis'
    settings['simulation']['zero_ensemble'] = True
    settings['simulation']['zero_left'] = -16.0
    settings['simulation']['flux'] = True
    settings['retis'] = {'rgen': 'rgen'}
    with patch('sys.stdout', new=StringIO()):
        fill_up_tis_and_retis_settings(settings)
        ensembles = create_ensembles(settings)
    return settings, ensembles


def create_variables1(exe_path=None):
    """Create system, engine, order parameter etc."""
    settings = {key: copy.deepcopy(val) for key, val in SETTINGS1.items()}
    add_default_settings(settings)
    settings['particles']['position']['input_file'] =\
        os.path.join(HERE, 'initial.xyz')
    settings['particles']['type'] = 'internal'
    settings['simulation']['task'] = 'retis'
    settings['simulation']['zero_ensemble'] = True
    settings['simulation']['flux'] = True
    settings['simulation']['exe_path'] = exe_path

    with patch('sys.stdout', new=StringIO()):
        fill_up_tis_and_retis_settings(settings)
        ensembles = create_ensembles(settings)
    return settings, ensembles


class TestPathSimulation(unittest.TestCase):
    """Run the tests for the PathSimulation class."""

    def test_init(self):
        """Test that we can create the simulation object."""
        settings, ensembles = create_variables()
        simulation = PathSimulation(ensembles, settings, controls={})
        self.assertTrue(hasattr(simulation, 'rgen'))
        settings['engine']['sigma_v'] = 0.1
        simulation = PathSimulation(ensembles, settings, controls={})
        self.assertEqual(len(simulation.ensembles), len(ensembles))
        for i, j in zip(simulation.ensembles, ensembles):
            self.assertIs(i, j)
        self.assertTrue(simulation.settings['tis']['aimless'])
        settings['simulation']['task'] = 'make-tis-files'
        del settings['ensemble']
        fill_up_tis_and_retis_settings(settings)
        for ens_set in settings['ensemble']:
            self.assertEqual('tis', ens_set['simulation']['task'])
        del settings['tis']
        with self.assertRaises(ValueError):
            PathSimulation(ensembles, settings, controls={})

    def test_init_sigma_v(self):
        """Test that we can create the simulation object."""
        settings, ensembles = create_variables()
        settings['ensemble'][1]['tis']['sigma_v'] = 0.1
        simulation = PathSimulation(ensembles, settings, controls={})
        self.assertFalse(simulation.settings['ensemble'][1]['tis']['aimless'])

    def test_restart_info(self):
        """Test generation of restart info."""
        settings, ensembles = create_variables()
        simulation = PathSimulation(ensembles,
                                    settings,
                                    controls={'steps': 16})
        info = simulation.restart_info()
        for key in ('cycle', 'rgen'):
            self.assertIn(key, info['simulation'])
        self.assertEqual(info['simulation']['rgen']['state'][2], 624)
        self.assertIn('system', info)
        simulation = PathSimulation(ensembles, settings, controls={})
        simulation.load_restart_info(info)
        self.assertEqual(simulation.rgen.get_state()['state'][2], 624)
        self.assertEqual(simulation.cycle['startcycle'], 16)

    def test_write_restart_file(self):
        """Test that we can create the restart files."""
        with tempfile.TemporaryDirectory() as tempdir:
            settings, ensembles = create_variables(exe_path=tempdir)
            simulation = PathSimulation(ensembles, settings, controls={})
            simulation.set_up_output(settings)
            simulation.write_restart(now=True)
            dirs = [i.name for i in os.scandir(tempdir) if i.is_dir()]
            self.assertEqual(len(dirs), 4)
            for i in ensembles:
                self.assertIn(i['path_ensemble'].ensemble_name_simple, dirs)
                for j in os.scandir(
                        i['path_ensemble'].directory['path_ensemble']):
                    if j.is_file():
                        self.assertEqual('ensemble.restart', j.name)
            files = [i.name for i in os.scandir(tempdir) if i.is_file()]
            self.assertEqual(len(files), 1)
            self.assertIn('pyretis.restart', files)

    def test_initiation(self):
        """Test the initiation method."""
        with tempfile.TemporaryDirectory() as tempdir:
            # Generate a input file in the temp directory
            with open(os.path.join(tempdir, 'initial.xyz'), 'w',
                      encoding='utf-8') as scrivi:
                scrivi.write('1\n \n Ar  -1.0  -0.0  -0.0 ')

            settings, ensembles = create_variables1(exe_path=tempdir)
            simulation = PathSimulation(ensembles, settings, controls={})
            simulation.set_up_output(settings)
            with patch('sys.stdout', new=StringIO()):
                self.assertTrue(simulation.initiate(settings))
            # Check the soft exit option:
            open(os.path.join(tempdir, 'EXIT'), 'w', encoding='utf-8').close()
            with patch('sys.stdout', new=StringIO()):
                self.assertFalse(simulation.initiate(settings))


class TestSimulationRETIS(unittest.TestCase):
    """Run the tests for the SimulationRETIS class."""

    def test_step(self):
        """Test the initiation method."""
        moves = {'move-0': 'tr', 'move-1': 'tr', 'move-2': 'tr'}

        with tempfile.TemporaryDirectory() as tempdir:
            settings, ensembles = create_variables1(exe_path=tempdir)
            simulation = SimulationRETIS(ensembles, settings, controls={})
            simulation.set_up_output(settings)
            with patch('sys.stdout', new=StringIO()):
                self.assertTrue(simulation.initiate(settings))
                result = simulation.step()
                for key, move in moves.items():
                    self.assertIn(key, result)
                    self.assertEqual(move, result[key])

    def test_soft_exit(self):
        """Test the soft exit method."""
        with tempfile.TemporaryDirectory() as tempdir:
            settings, ensembles = create_variables1(exe_path=tempdir)
            simulation = SimulationRETIS(ensembles, settings, controls={})
            settings['simulation']['exe_path'] = tempdir
            simulation.set_up_output(settings)
            simulation.cycle['endcycle'] = 1
            exit_file = os.path.join(tempdir, 'EXIT')
            with patch('sys.stdout', new=StringIO()) as stdout:
                self.assertTrue(simulation.initiate(settings))
                pathlib.Path(exit_file).touch()
                for _ in enumerate(simulation.run()):
                    pass
                self.assertIn('soft exit', stdout.getvalue().strip())
                restart_file = os.path.join(tempdir, 'pyretis.restart')
                self.assertTrue(os.path.exists(restart_file))
                self.assertEqual(simulation.cycle['step'], 1)

    def test_priority(self):
        """Test the priority shooting feature."""
        with tempfile.TemporaryDirectory() as tempdir:
            settings, ensembles = create_variables1(exe_path=tempdir)
            fill_up_tis_and_retis_settings(settings)
            simulation = SimulationRETIS(ensembles, settings, controls={})
            simulation.set_up_output(settings)
            with patch('sys.stdout', new=StringIO()):
                self.assertTrue(simulation.initiate(settings))
                simulation.cycle['step'] = 2
                settings['retis']['swapfreq'] = 0.0
                settings['simulation']['priority_shooting'] = True
                simulation.ensembles[0]['path_ensemble'].nstats['npath'] = 2
                simulation.ensembles[1]['path_ensemble'].nstats['npath'] = 1
                simulation.ensembles[2]['path_ensemble'].nstats['npath'] = 1

                # Get nstats['npath'] = 2 for all path_ensembles:
                result = simulation.step()
                self.assertNotIn('move-0', result)
                self.assertIn('move-1', result)
                self.assertIn('move-2', result)
                self.assertEqual([2, 2, 2], [i['path_ensemble'].nstats['npath']
                                             for i in simulation.ensembles])

                # test case when we 100% do swap with unequal nstats['npath']:
                simulation.ensembles[2]['path_ensemble'].nstats['npath'] = 0
                settings['retis']['swapfreq'] = 1.0

                # 1. We do TIS move in ensembles[2] only:
                result = simulation.step()
                self.assertNotIn('move-0', result)
                self.assertIn('move-2', result)

                # We do swap move in when priority_shooting = False:
                settings['simulation']['priority_shooting'] = False
                result = simulation.step()
                for i in range(3):
                    move = 'swap' if i else 'nullmove'
                    self.assertEqual(move, result.get('move-' + str(i)))

    def test_priority_tis(self):
        """Test the priority shooting feature for tis."""
        with tempfile.TemporaryDirectory() as tempdir:
            settings, _ = create_variables1(exe_path=tempdir)
            settings['simulation']['task'] = 'tis'
            settings['simulation']['interfaces'] = [-0.9, -0.8, -0.7, 1.0]
            settings['simulation']['flux'] = False
            del settings['ensemble']
            fill_up_tis_and_retis_settings(settings)
            with patch('sys.stdout', new=StringIO()):
                ensembles = create_ensembles(settings)
                simulation = SimulationTIS(ensembles, settings, controls={})
                simulation.set_up_output(settings)
                self.assertTrue(simulation.initiate(settings))
                simulation.cycle['endcycle'] = 4
                simulation.cycle['step'] = 2
                settings['simulation']['priority_shooting'] = True
                simulation.ensembles[0]['path_ensemble'].nstats['npath'] = 2
                simulation.ensembles[1]['path_ensemble'].nstats['npath'] = 1
                simulation.ensembles[2]['path_ensemble'].nstats['npath'] = 1

                cnt = 0
                while not simulation.is_finished():
                    simulation.step()
                    cnt += 1
                self.assertEqual([4, 4, 4], [i['path_ensemble'].nstats['npath']
                                             for i in simulation.ensembles])
                self.assertEqual(cnt, 3)
                self.assertTrue(simulation.is_finished())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the trajectory writers."""
import logging
import unittest
import tempfile
import os
import numpy as np
from numpy.random import rand, random
from pyretis.core import create_box, System, Particles, Path
from pyretis.tools.lattice import generate_lattice
from pyretis.inout.formats.snapshot import (
    SnapshotFile,
    SnapshotFormatter,
)
from pyretis.inout.formats.path import (
    PathIntFormatter,
    PathExtFormatter,
    PathExtFile,
    PathIntFile,
)
from .help import create_external_path, CORRECT_PATH_EXT, set_up_system
logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


def create_test_system():
    """Create a system we can use for testing."""
    xyz, size = generate_lattice('fcc', [3, 3, 3], density=0.9)
    low, high = [], []
    for i in size:
        low.append(i[0])
        high.append(i[1])
    box = create_box(low=low, high=high)
    system = System(units='lj', box=box)
    system.particles = Particles(dim=3)
    for xyzi in xyz:
        system.add_particle(name='Ar', pos=xyzi, vel=np.zeros_like(xyzi))
    system.particles.vel[-1] = np.array([-1.0, 0.123, 1.0])
    return system


def create_path():
    """Setup a simple path for a test."""
    box = create_box(cell=rand(3))
    path = Path(None)
    phasepoints = []
    for _ in range(10):
        pos = rand(10, 3)
        vel = rand(*pos.shape)
        vpot = random()
        ekin = random()
        phasepoint = set_up_system(
            [pos[0][0], pos[1][0]], pos, vel, vpot=vpot, ekin=ekin
        )
        phasepoint.box = box
        phasepoints.append(phasepoint)
        path.append(phasepoint)
    return phasepoints, path


class TrajTest(unittest.TestCase):
    """Test trajectory writing work as intended."""

    def test_txt_writer(self):
        """Test the SnapshotFormatter."""
        system = create_test_system()
        txt_writer = SnapshotFormatter(write_vel=True)
        snapshot = txt_writer.format_snapshot(0, system)
        correct = os.path.join(HERE, 'generated.txt')
        with open(correct, 'r', encoding='utf-8') as fileh:
            for lines1, lines2 in zip(fileh, snapshot):
                self.assertEqual(lines1.rstrip(), lines2.rstrip())

    def test_traj_writer_novel(self):
        """Test the SnapShotFormatter class when we exclude velocities."""
        writer = SnapshotFormatter(write_vel=False, fmt='full')
        phasepoints, path = create_path()
        with tempfile.NamedTemporaryFile() as tmp:
            for step, snapshot in enumerate(path.phasepoints):
                for line in writer.format_snapshot(step, snapshot):
                    string = f'{line}\n'
                    tmp.write(string.encode('utf-8'))
            tmp.flush()
            del writer
            reader = SnapshotFile(tmp.name, 'r')
            for block, snapshot in zip(reader.load(), phasepoints):
                self.assertTrue(np.allclose(block['box'], snapshot.box.length))
                xyz = np.transpose(np.vstack((block['x'],
                                              block['y'],
                                              block['z'])))
                self.assertTrue(np.allclose(xyz, snapshot.particles.get_pos()))

    def test_traj_writer_vel(self):
        """Test the SnapshotFormatter class when we include velocities."""
        writer = SnapshotFormatter(write_vel=True, fmt='full')
        phasepoints, path = create_path()
        with tempfile.NamedTemporaryFile() as tmp:
            for step, snapshot in enumerate(path.phasepoints):
                for line in writer.format_snapshot(step, snapshot):
                    string = f'{line}\n'
                    tmp.write(string.encode('utf-8'))
            tmp.flush()
            del writer
            reader = SnapshotFile(tmp.name, 'r')
            for block, snapshot in zip(reader.load(), phasepoints):
                self.assertTrue(np.allclose(block['box'], snapshot.box.length))
                xyz = np.transpose(np.vstack((block['x'],
                                              block['y'],
                                              block['z'])))
                vel = np.transpose(np.vstack((block['vx'],
                                              block['vy'],
                                              block['vz'])))
                self.assertTrue(np.allclose(xyz, snapshot.particles.get_pos()))
                self.assertTrue(np.allclose(vel, snapshot.particles.get_vel()))

    def test_path_int_writer(self):
        """Test the path internal writer."""
        phasepoints, path = create_path()
        writer = PathIntFormatter()
        fmt = ' '.join([writer._FMT] * 6)
        idxs = 0
        idx = 0
        for i, lines in enumerate(writer.format(0, (path, 'ACC'))):
            if i == 0:
                self.assertEqual('# Cycle: 0, status: ACC', lines)
            else:
                if lines.startswith('Snapshot'):
                    self.assertEqual(f'Snapshot: {idxs}', lines)
                    idxs += 1
                    idx = 0
                else:
                    posvel = fmt.format(
                        *phasepoints[idxs - 1].particles.get_pos()[idx],
                        *phasepoints[idxs - 1].particles.get_vel()[idx]
                    )
                    self.assertEqual(lines, posvel)
                    idx += 1

    def test_pathint_read_write(self):
        """Test that we can read write with PathIntFormatter."""
        writer = PathIntFormatter()
        statuses = ('ACC', 'BWI', 'FTL')
        path_phasepoints = []
        with tempfile.NamedTemporaryFile() as tmp:
            for i, status in zip(range(3), statuses):
                phasepoints, path = create_path()
                path_phasepoints.append(phasepoints)
                for line in writer.format(i, (path, status)):
                    tmp.write(f'{line}\n'.encode('utf-8'))
            tmp.flush()
            del writer
            reader = PathIntFile(tmp.name, 'r')
            for status, path, phasepoints in zip(statuses,
                                                 reader.load(),
                                                 path_phasepoints):
                self.assertEqual(status, path['comment'][0].split()[-1])
                for snapshot, phasepoint in zip(path['data'], phasepoints):
                    particles = phasepoint.particles
                    self.assertTrue(
                        np.allclose(snapshot['pos'], particles.get_pos())
                    )
                    self.assertTrue(
                        np.allclose(snapshot['vel'], particles.get_vel())
                    )

    def test_pathint_read_error(self):
        """Test what happens in we read faulty input with PathIntFile."""
        filename = os.path.join(HERE, 'traj-error1.txt')
        reader = PathIntFile(filename, 'r').load()
        next(reader)
        with self.assertRaises(ValueError):
            next(reader)
        filename = os.path.join(HERE, 'traj-error2.txt')
        reader = PathIntFile(filename, 'r').load()
        with self.assertRaises(ValueError):
            next(reader)

    def test_path_ext_writer(self):
        """Test the path external writer."""
        path, _ = create_external_path()
        writer = PathExtFormatter()
        for corr, snap in zip(CORRECT_PATH_EXT,
                              writer.format(0, (path, 'ACC'))):
            self.assertEqual(corr, snap)

    def test_path_ext_read_write(self):
        """Test the read/write for the PathExtFormatter."""
        writer = PathExtFormatter()
        statuses = ('ACC', 'BWI', 'FTL')
        all_path_data = []
        with tempfile.NamedTemporaryFile() as tmp:
            for i, status in zip(range(3), statuses):
                path, data = create_external_path(random_length=True)
                all_path_data.append(data)
                for line in writer.format(i, (path, status)):
                    string = f'{line}\n'
                    tmp.write(string.encode('utf-8'))
            tmp.flush()
            del writer
            reader = PathExtFile(tmp.name, 'r')
            for status, path, data in zip(statuses, reader.load(),
                                          all_path_data):
                self.assertEqual(status, path['comment'][0].split()[-1])
                for snapshot, datai in zip(path['data'], data):
                    self.assertEqual(int(snapshot[0]), datai[0])
                    self.assertEqual(snapshot[1], datai[1])
                    if datai[2] is None:
                        self.assertEqual(int(snapshot[2]), 0)
                    else:
                        self.assertEqual(int(snapshot[2]), datai[2])
                    vel = snapshot[3] == '-1'
                    self.assertEqual(vel, datai[3])

    def test_path_eq_none(self):
        """
        If nullmoves=False, the first and last ensembles are sometimes not
        updated (in first and last path ensembles when they are excluded from
        a replica exchange move) and nothing is supposed to be written to file.
        This is tested here.
        """
        writer = PathExtFormatter()
        self.assertEqual(list(writer.format(4, (None, 'ACC'))), [])
        writer = PathIntFormatter()
        self.assertEqual(list(writer.format(13, (None, 'ACC'))), [])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the bin."""
import logging
import os
import unittest
import shutil
import tempfile
import subprocess
from io import StringIO
from unittest.mock import patch
from pyretis.bin.pyretisanalyse import (hello_world,
                                        bye_bye_world,
                                        main)

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class test_pyretis_main_analyse(unittest.TestCase):
    """Test Main."""

    def test_pyretisanalyse(self):
        """Test pyretisanalyse."""
        with tempfile.TemporaryDirectory() as tempdir:
            with patch('sys.stdout', new=StringIO()):
                with subprocess.Popen(['pyretisanalyse', '-V'],
                                      stderr=subprocess.DEVNULL,
                                      stdout=subprocess.PIPE) as ex:
                    asd = ex.stdout.read().split()
                self.assertTrue(b'PyRETIS' in asd)

            with patch('sys.stdout', new=StringIO()) as stdout:
                with subprocess.Popen(['pyretisanalyse'],
                                      stderr=subprocess.DEVNULL,
                                      stdout=subprocess.PIPE,
                                      cwd=tempdir) as ex:
                    asd = ex.stdout.read().strip()
                self.assertIn(b'Input file required', asd)

            inputfile = 'a_non_existent_input.rst'
            with patch('sys.stdout', new=StringIO()) as stdout:
                main(inputfile, tempdir, tempdir)
            self.assertIn('NOT found!', stdout.getvalue().strip())

            with patch('sys.stdout', new=StringIO()) as stdout:
                main(None, tempdir, tempdir)
            self.assertIn('Input file required', stdout.getvalue().strip())

    def test_hello_world(self):
        """Test that we are polite."""
        with patch('sys.stdout', new=StringIO()) as stdout:
            hello_world(infile='I_can_read_your_soul.rst',
                        run_dir='You_are_wasting_your_life_here',
                        report_dir='go_out_and_enjoy')
        self.assertIn('Start', stdout.getvalue().strip())

    def test_bye_world(self):
        """Test that we can die."""
        with patch('sys.stdout', new=StringIO()) as stdout:
            bye_bye_world()
        self.assertIn('reference', stdout.getvalue().strip())

    def test_main_an(self):
        """Test that we know how to set up a report."""
        input_file = 'settings-retis.rst'
        # fake pathensemble line
        lines = ['']*8
        moves = ['ki', 'sh', 'ki', 'sh', 'ld', 'xx', 'sh']
        line = "         2          3          0 L M R      99 ACC "
        line += "sh -9.108621358e-02  1.007510172e+01       0"
        line += "      98  0.000000000e+00       0       0  1.00e+00\n"
        for j in range(7):
            lines[j] = line.replace('sh', moves[j])
        lines[7] = line.replace('ACC', 'BWI')
        lines[0] = line.replace('L', 'R')
        lines[1] = line.replace('L', 'R')
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copyfile(os.path.join(HERE, input_file),
                            os.path.join(tempdir, input_file))
            with patch('sys.stdout', new=StringIO()) as stdout:
                main(os.path.join(tempdir, input_file), tempdir,
                     os.path.join(tempdir, 'report'))
            self.assertIn('traceback', stdout.getvalue().strip())

            with patch('sys.stdout', new=StringIO()) as stdout:
                for i in ['000', '001', '002']:
                    i_folder = os.path.join(tempdir, i)
                    os.mkdir(i_folder)
                    with open(i_folder + '/pathensemble.txt', 'w') as f:
                        for line in lines:
                            f.write(line)
                i_folder = os.path.join(tempdir, '003')
                os.mkdir(i_folder)
                with open(i_folder + '/pathensemble.txt', 'w') as f:
                    for line in lines[:-2]:
                        f.write(line)

                main(os.path.join(tempdir, input_file), tempdir,
                     os.path.join(tempdir, 'report'))

            self.assertIn('ssing probability: 1.00', stdout.getvalue().strip())
            self.assertIn('2.577319588', stdout.getvalue().strip())
            input_file = 'dummy_input.rst'  # TIS
            shutil.copyfile(os.path.join(HERE, input_file),
                            os.path.join(tempdir, input_file))
            with patch('sys.stdout', new=StringIO()) as stdout:
                main(os.path.join(tempdir, input_file), tempdir,
                     os.path.join(tempdir, 'report'))
            self.assertIn('report/002_tis_report.', stdout.getvalue().strip())

    def test_main_repptis(self):
        """Test pyretisanalyse (main) for a REPPTIS simulation.

        This one, just like the pyretisanalyse test for RETIS simuls,
        takes quite a long time because the output is zipped, and
        figures are created...

        """
        input_file = os.path.join("pptis", "retis.rst")
        pathensemblefiles =\
            [os.path.join("pptis",
                          "pathensemble"+str(i)+".txt") for i in range(4)]
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copyfile(os.path.join(HERE, input_file),
                            os.path.join(tempdir, "retis.rst"))
            for i in range(4):
                os.mkdir(os.path.join(tempdir, "00"+str(i)))
                shutil.copyfile(os.path.join(HERE, pathensemblefiles[i]),
                                os.path.join(tempdir, "00"+str(i),
                                             "pathensemble.txt"))
            with patch('sys.stdout', new=StringIO()) as stdout:
                main(os.path.join(tempdir, "retis.rst"), tempdir,
                     os.path.join(tempdir, 'report'))
            self.assertIn('ssing probability: 5.166013535e-04',
                          stdout.getvalue().strip())
            self.assertIn('reduced): 0.734638269', stdout.getvalue().strip())
            self.assertIn('Xi: 0.46000', stdout.getvalue().strip())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Define a new directive for PyRETIS keywords."""
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged
from jinja2 import Template


HTML_PRE = """
<div class="container">
    <div class="panel panel-keyword">
      <div class="panel-heading">{{ heading }}</div>
      <div class="panel-body">"""
TEMPLATE_PRE = Template(HTML_PRE)
HTML_POST = """</div>
  </div>
</div>"""
TEMPLATE_POST = Template(HTML_POST)


class PyretisKeywordNode(nodes.General, nodes.Element):
    pass


class PyretisKeyword(Directive):

    required_arguments = 2
    optional_arguments = 5
    option_spec = {'specific': unchanged}
    has_content = True

    def run(self):
        node = PyretisKeywordNode()
        if len(self.arguments) == 2:
            arguments = [i for i in self.arguments]
        else:
            if self.arguments[0].lower() == 'parameter':
                arguments = [
                    '{} {}'.format(
                        self.arguments[0],
                        self.arguments[1]
                    )
                ]
                arguments += [' '.join(self.arguments[2:])]
            else:
                arguments = ([self.arguments[0]] +
                             [' '.join(self.arguments[1:])])
        if self.options.get('specific', 'no') == 'yes':
            node['heading'] = '{} = {}'.format(*arguments)
        else:
            node['heading'] = '{} = <i>{}</i>'.format(*arguments)
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


def html_visit_keyword_block(self, node):
    self.body.append(
        TEMPLATE_PRE.render(
            heading=node['heading']
        )
    )


def html_depart_keyword_block(self, node):
    self.body.append(TEMPLATE_POST.render())


def setup(app):
    app.add_node(
        PyretisKeywordNode,
        html=(
            html_visit_keyword_block,
            html_depart_keyword_block
        )
    )
    app.add_directive('pyretis-keyword', PyretisKeyword)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A script for building the C extension."""
from setuptools import (
    setup,
    Extension,
)
import numpy as np


LJMODULE = Extension(
    'ljc',
    sources=['ljc.c'],
    extra_compile_args=['-Ofast', '-march=native'],
)
setup(
    name='PyRETIS Lennard-Jones c extension',
    description='C extension for the Lennard-Jones potential.',
    ext_modules=[LJMODULE],
    include_dirs=[np.get_include()],
)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Define common methods and variables for the tests"""
from contextlib import contextmanager
from pyretis.inout.formats.formatter import OutputFormatter
import logging


@contextmanager
def turn_on_logging():
    """Turn on logging so that tests can detect it."""
    logging.disable(logging.NOTSET)
    try:
        yield
    finally:
        logging.disable(logging.CRITICAL)


TEST_SETTINGS = {
    'system': {
        'dimensions': 3,
        'units': 'reduced',
        'temperature': 1.0
    },
    'particles': {
        'position': {
            'generate': 'fcc',
            'repeat': [2, 2, 2],
            'density': 0.9,
        }
    },
    'potential': [
        {
            'class': 'PairLennardJonesCutnp',
            'shift': True,
            'parameter': {
                0: {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 2.5}
            }
        }
    ]
}


class TxtWriter(OutputFormatter):
    """A class used for testing output."""
    FMT = '{:>10d} {:>20s}'

    def format(self, step, data):
        yield self.FMT.format(step, data)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the common methods in pyretis.inout.common."""
import os
import logging
import unittest
from pyretis.inout.common import (
    make_dirs,
    simplify_ensemble_name,
    add_dirname,
    name_file,
    generate_file_name,
    create_empty_ensembles
)


logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


def remove_dir(dirname):
    """Remove a directory."""
    try:
        os.removedirs(dirname)
    except OSError:
        pass


class TestMethods(unittest.TestCase):
    """Test some of the methods from pyretis.inout.common."""
    def test_create_empty_ensembles(self):
        """Test that we can properly create empty ensembles."""
        settings = {'simulation': {'interfaces': [1, 2, 3, 4, 5],
                                   'flux': True,
                                   'zero_ensemble': True}}
        create_empty_ensembles(settings)
        settings = {'simulation': {'interfaces': [1, 2, 3, 4, 5],
                                   'flux': True,
                                   'zero_ensemble': True}}
        settings['ensemble'] = [settings]
        settings['ensemble'][0]['interface'] = 1
        settings['ensemble'][0]['tis'] = {'ensemble_number': 123321}
        create_empty_ensembles(settings)
        self.assertTrue(settings['ensemble'][0]['interface'] == 1)
        self.assertTrue(
            settings['ensemble'][0]['tis']['ensemble_number'] == 123321)

        settings = {'simulation': {'interfaces': [1, 2, 3, 4, 5],
                                   'flux': True,
                                   'zero_ensemble': True},
                    'tis': {'ensemble_number': 123321}}
        create_empty_ensembles(settings)
        self.assertTrue(
            settings['ensemble'][0]['tis']['ensemble_number'] == 123321)
        self.assertTrue(len(settings['ensemble']) == 1)

        settings = {'simulation': {'interfaces': [1, 2, 3, 4, 5],
                                   'flux': True,
                                   'zero_ensemble': True},
                    'ensemble': [{'interface': 3,
                                  'gnappo': 'lappo',
                                  'tis': {'ensemble_number': 3}}]}
        create_empty_ensembles(settings)
        self.assertEqual(settings['ensemble'][3]['interface'], 3)
        self.assertEqual(settings['ensemble'][3]['gnappo'], 'lappo')
        self.assertEqual(
            settings['ensemble'][3]['tis']['ensemble_number'], 3)
        # todo. This should be part of the ensamble numering issue.
        # with self.assertRaises(ValueError) as err:
        #    create_empty_ensembles(settings)
        # self.assertEqual('Invalid entry for setting ensemble',
        #    str(err.exception))
        # settings['ensemble'][0] = {'tis': {'ensemble_number': 'flux'}}
        # create_empty_ensembles(settings)
        # self.assertTrue(
        #     settings['simulation']['ensemble'][0]['tis']['ensemble_number'],
        #     'flux')

    def test_make_dirs(self):
        """Test that we can create directories."""
        dirname = os.path.join(HERE, 'testdir')
        make_dirs(dirname)
        self.assertTrue(os.path.isdir(dirname))
        remove_dir(dirname)
        dirname = os.path.join(HERE, 'already_exists')
        with self.assertRaises(OSError):
            make_dirs(dirname)
        dirname = os.path.join(HERE, 'dir_exists')
        remove_dir(dirname)
        make_dirs(dirname)
        msg = make_dirs(dirname)
        self.assertTrue(msg.endswith('already exist.'))
        remove_dir(dirname)

    def test_simplify_ensemble_name(self):
        """Test that we can simplify ensemble names."""
        cases = [('[0^-]', '000'), ('[0^+]', '001'), ('[1^+]', '002')]
        for case in cases:
            txt = simplify_ensemble_name(case[0], fmt='{:03d}')
            self.assertEqual(txt, case[1])
        txt = simplify_ensemble_name('001', fmt='{:03d}')
        self.assertEqual(txt, '001')
        txt = simplify_ensemble_name('[1]', fmt='{:03d}')
        self.assertEqual(txt, '002')

    def test_add_dirname(self):
        """Test that we can add a directory to a filename."""
        path = add_dirname('filename.txt', 'path')
        self.assertEqual(path, os.path.join('path', 'filename.txt'))
        path = add_dirname('filename.txt', None)
        self.assertEqual(path, 'filename.txt')

    def test_name_file(self):
        """Test that we can name a file."""
        name = name_file('test', 'txt', path='path')
        filename = ''.join(['test', os.extsep, 'txt'])
        filepath = os.path.join('path', filename)
        self.assertEqual(name, filepath)

    def test_generate_filename(self):
        """Test the generation of file names."""
        settings = {'output': {}}
        name = generate_file_name('base.txt', 'dir', settings)
        self.assertEqual(name, os.path.join('dir', 'base.txt'))
        settings = {'output': {'prefix': 'abc-'}}
        name = generate_file_name('base.txt', 'dir', settings)
        self.assertEqual(name, os.path.join('dir', 'abc-base.txt'))

    def test_check_python_version(self):
        # DO custom import and go from there to prevent overriding anything
        # important
        # Set the min python version to next mayor version to remember updating
        import pyretis.inout.common as common
        common.MIN_PYTHON_VERSION = (4, 0)
        with self.assertRaises(SystemExit):
            common.check_python_version()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Plot raw data from a simulation."""
# pylint: disable=invalid-name
import numpy as np
import colorama
from matplotlib import pyplot as plt
from matplotlib.widgets import Button, RadioButtons
from matplotlib.cm import get_cmap
from pyretis.core import create_box, System, Particles
from pyretis.core.tis import shoot, time_reversal
from pyretis.core.retis import retis_swap
from pyretis.initiation import initiate_path_simulation
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.setup import create_simulation, create_force_field


def plot_path(path, axi, axj, color, alphai=0.9, alphaj=0.9, lsj='-'):
    """Plot data for paths in ensembles."""
    pos = []
    order = []
    time = []
    for i, phasepoint in enumerate(path.phasepoints):
        pos.append(phasepoint.particles.get_pos()[0])
        order.append(phasepoint.order[0])
        time.append(i)
    pos = np.array(pos)
    order = np.array(order)
    axi.plot(pos[:, 0], pos[:, 1], lw=3, alpha=alphai,
             color=color)
    axj.plot(time, order, lw=3, alpha=alphaj, color=color, ls=lsj)
    axi.scatter(pos[0, 0], pos[0, 1], s=50, marker='x', color=color,
                alpha=alphai)
    axj.scatter(time[0], order[0], s=50, marker='x', color=color,
                alpha=alphaj)
    if order[-1] < order[-2]:
        end = '^'
    else:
        end = 'v'
    axi.scatter(pos[-1, 0], pos[-1, 1], s=50, marker=end, color=color,
                alpha=alphai)
    axj.scatter(time[-1], order[-1], s=50, marker=end, color=color,
                alpha=alphaj)


class PlotHelper:
    """A class for handling the plotting."""

    def __init__(self, simulation, settings, axi, axj, colors):
        """Set up the plotter."""
        self.simulation = simulation
        self.axi = axi
        self.axj = axj
        self.colors = colors
        self.potential_setup(settings)
        self.update = None
        self.last_result = None
        self.idx = 0

    def potential_setup(self, settings):
        """Set up the potential function."""
        # Set up raw data:
        forcefield = create_force_field(settings)
        box = create_box(periodic=[False, False])
        fakesys = System(units='reduced', box=box)
        fakesys.particles = Particles(dim=2)
        fakesys.add_particle(name='B', pos=np.zeros(2), ptype=1)
        minx, maxx = -0.5, 0.5
        miny, maxy = -1.0, 1.0
        xval = np.linspace(minx, maxx, 100)
        yval = np.linspace(miny, maxy, 100)
        xpos, ypos = np.meshgrid(xval, yval, indexing='ij')
        pot = np.zeros_like(xpos)
        for i, x in enumerate(xval):
            for j, y in enumerate(yval):
                fakesys.particles.pos[0, 0] = x
                fakesys.particles.pos[0, 1] = y
                pot[i, j] = forcefield.evaluate_potential(fakesys)
        self.pot = pot
        self.xpos = xpos
        self.ypos = ypos
        self.interfaces = settings['simulation']['interfaces']
        self.inter_a = settings['orderparameter']['inter_a']
        self.inter_b = settings['orderparameter']['inter_b']
        self.minx = minx
        self.maxx = maxx
        self.miny = miny
        self.maxy = maxy

    def plot_potential(self, axi, axj):
        """Plot the potential in the given axis."""
        axi.contourf(self.xpos, self.ypos, self.pot, 10,
                     cmap=get_cmap('viridis'), alpha=0.8)
        # Add interfaces:
        for inter in self.interfaces:
            axi.axhline(y=inter, lw=2, ls=':', color='#262626', alpha=0.8)
            axj.axhline(y=inter, lw=2, ls=':', color='#262626', alpha=0.8)
        extra_int = [self.inter_a, self.inter_b]
        for inter in extra_int:
            axi.axhline(y=inter, lw=2, ls=':', alpha=0.8, color='#BAACE6')
            axj.axhline(y=inter, lw=2, ls=':', alpha=0.8, color='#BAACE6')
        axi.set_xlim((self.minx, self.maxx))
        axi.set_ylim((self.miny, self.maxy))
        axi.set_xlabel((r'Position ($x$)'), fontsize='large')
        axi.set_ylabel((r'Position ($y$)'), fontsize='large')
        axj.set_xlabel((r'Step number'), fontsize='large')
        axj.set_ylabel((r'Order parameter ($\lambda$)'), fontsize='large')

    def plot_accepted(self, idx=None):
        """Plot the accepted paths."""
        ensembles = self.simulation.ensembles
        for i, ensemble in enumerate(ensembles):
            if idx is None:
                alpha = 0.9
            else:
                if i in idx:
                    alpha = 0.9
                else:
                    alpha = 0.25
            plot_path(ensemble['path_ensemble'].last_path, self.axi, self.axj,
                      self.colors[i], alphai=0.9, alphaj=alpha)

    def do_shoot(self, event):
        """Perform shooting moves."""
        ensembles = self.simulation.ensembles
        ensemble = ensembles[self.idx]
        accept, trial, status = shoot(
            ensemble,
            self.simulation.settings['tis'],
            ensemble['path_ensemble'].start_condition
        )
        self.axi.clear()
        self.axj.clear()
        self.plot_potential(self.axi, self.axj)
        self.plot_accepted((self.idx,))
        plot_path(trial, self.axi, self.axj, self.colors[-1], lsj='--')
        sidx = trial.generated[2]
        spoint = ensemble['path_ensemble'].last_path.phasepoints[sidx]
        pos = spoint.particles.get_pos()[0]
        order = spoint.order[0]
        self.axi.scatter(pos[0], pos[1], marker='o', s=75, alpha=0.9,
                         color=self.colors[self.idx])
        self.axj.scatter(sidx, order, marker='o', s=75, alpha=0.9,
                         color=self.colors[self.idx])
        self.last_result = ('Shoot', ensemble, accept, trial, status, self.idx)
        self.axj.relim()
        self.axj.autoscale_view()
        plt.draw()

    def do_timereversal(self, event):
        """Perform time reversal."""
        ensembles = self.simulation.ensembles
        ensemble = ensembles[self.idx]
        accept, trial, status = time_reversal(
            ensemble['path_ensemble'].last_path,
            ensemble['order_function'],
            ensemble['path_ensemble'].interfaces,
            ensemble['path_ensemble'].start_condition
        )
        self.axi.clear()
        self.axj.clear()
        self.plot_potential(self.axi, self.axj)
        self.plot_accepted((self.idx,))
        plot_path(trial, self.axi, self.axj, self.colors[-1], lsj='--')
        self.last_result = ('Time-reversal', ensemble, accept, trial,
                            status, self.idx)
        self.axj.relim()
        self.axj.autoscale_view()
        plt.draw()

    def do_accept_last_move(self, event):
        """Try to accept the last move."""
        self.accept_last_move()

    def accept_last_move(self):
        """Check if we can accept the last move."""
        if self.last_result is not None:
            move, ensemble, accept, trial, status, _ = self.last_result
            if move == 'Swap':
                # idxs = (idx, idx + 1)
                print_to_screen(
                    'Swap: {} <-> {}'.format(
                        ensemble[0]['path_ensemble'].ensemble_name,
                        ensemble[1]['path_ensemble'].ensemble_name
                    ),
                    level='message'
                )
            else:
                # idxs = (idx,)
                ensemble['path_ensemble'].add_path_data(trial, status)
                print_to_screen(
                    f"In ensemble: {ensemble['path_ensemble'].ensemble_name}",
                    level='message'
                )
            if accept:
                print_to_screen(f'\t{move} was accepted!',
                                level='success')
                print_to_screen(f'\t{status}', level='success')
            else:
                print_to_screen(f'\t{move} was rejected!',
                                level='error')
                print_to_screen(f'\t{status}', level='error')
            self.axi.clear()
            self.axj.clear()
            self.plot_potential(self.axi, self.axj)
            self.plot_accepted()
            self.last_result = None
            self.axj.relim()
            self.axj.autoscale_view()
            plt.draw()

    def set_idx(self, label):
        """Just to define indices for the ensembles."""
        idx = {'$[0^-]$': 0, '$[0^+]$': 1, '$[1^+]$': 2,
               '$[2^+]$': 3, '$[3^+]$': 4, '$[4^+]$': 5,
               '$[5^+]$': 6, '$[6^+]$': 7}
        self.idx = idx.get(label, 0)

    def do_swap(self, event):
        """Perform swapping when button is pressed."""
        if self.idx == len(self.simulation.ensembles) - 1:
            return
        accept, trial, status = retis_swap(
            self.simulation.ensembles,
            self.idx,
            self.simulation.settings,
            0
        )
        ensemble1 = self.simulation.ensembles[self.idx]
        ensemble2 = self.simulation.ensembles[self.idx + 1]
        self.last_result = ('Swap', (ensemble1, ensemble2), accept, trial,
                            status, self.idx)
        self.axi.clear()
        self.axj.clear()
        self.plot_potential(self.axi, self.axj)
        self.plot_accepted((self.idx, self.idx+1))
        self.axj.relim()
        self.axj.autoscale_view()
        plt.draw()


def set_up_simulation(settings):
    """Do all the set-ups to create the simulation."""
    simulation = create_simulation(settings)
    # Also, do the initialisation here:
    for i, _ in enumerate(initiate_path_simulation(simulation, settings)):
        ensemble = simulation.ensembles[i]['path_ensemble']
        name = ensemble.ensemble_name
        print_to_screen(f'Info about ensemble {name}:',
                        level='success')
        print_to_screen(ensemble)
        print_to_screen('Info about the initial path:', level='success')
        print_to_screen(ensemble.last_path)
        print_to_screen('')
    return simulation


def main(inputfile='retis.rst'):
    """Run the whole interactive plotting."""
    # Setup simulation:
    sim_settings = parse_settings_file(inputfile)
    simulation = set_up_simulation(sim_settings)
    # Set up for plotting:
    fig = plt.figure(figsize=(12, 6))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    axbt = fig.add_axes([0.20, 0.12, 0.1, 0.075])
    axbt2 = fig.add_axes([0.30, 0.12, 0.1, 0.075])
    axbt3 = fig.add_axes([0.40, 0.12, 0.1, 0.075])
    axbt4 = fig.add_axes([0.30, 0.04, 0.1, 0.075])

    cmap = get_cmap(name='tab10')
    colors = cmap(np.linspace(0, 1, len(simulation.ensembles)+2))

    myplot = PlotHelper(simulation, sim_settings, ax1, ax2, colors)

    # Add buttons:
    button1 = Button(axbt, 'Shoot')
    button1.on_clicked(myplot.do_shoot)
    button2 = Button(axbt2, 'Time reversal')
    button2.on_clicked(myplot.do_timereversal)
    button3 = Button(axbt3, 'Swap')
    button3.on_clicked(myplot.do_swap)
    button4 = Button(axbt4, 'Accept?')
    button4.on_clicked(myplot.do_accept_last_move)

    # Add radio button:
    rax = fig.add_axes([0.02, 0.2, 0.1, 0.6],
                       frameon=True)
    names = []
    for ensemble in simulation.ensembles:
        names.append(f"${ensemble['path_ensemble'].ensemble_name}$")
    radio = RadioButtons(rax, names, activecolor='#262626')
    for circle in radio.circles:
        circle.set_radius(0.02)
    for i, lab in enumerate(radio.labels):
        lab.set_fontsize(12)
        lab.set_color(colors[i])
    radio.on_clicked(myplot.set_idx)

    myplot.plot_potential(ax1, ax2)
    plt.subplots_adjust(right=0.95, left=0.20, top=0.95,
                        bottom=0.3, wspace=0.3)
    for i, ensemble in enumerate(simulation.ensembles):
        plot_path(ensemble['path_ensemble'].last_path, ax1, ax2, colors[i])
    plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main(inputfile='retis-y.rst')



# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains classes to represent order parameters.

The order parameters are assumed to all be completely determined
by the system properties and they will all return at least one
value - the order parameter itself. The order parameters can also
return several order parameters which can be used for further analysis.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Position (:py:class:`.Position`)
    An order parameter equal to the position of a particle.

AreaAndVolume (:py:class:`.AreaAndVolume`)
    An order parameter equal to the Area or Volume of a group of a particles.

"""
import logging
import numpy as np
from scipy.spatial import ConvexHull
from pyretis.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class AreaAndVolume(OrderParameter):
    """AreaAndVolume(OrderParameter).

    This order parameter calculates the area of the six-membered ring
    which the methane molecule jumps through when performing the L6L jump,
    and the volume of the starting cage.

    Attributes:
    ----------
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the position or not.

    """

    def __init__(self, periodic=True):
        """Initialize the class."""
        super().__init__(description="Area of ring and volume "
                                     "of starting cage")
        self.periodic = periodic
        self.idx1 = np.array([220, 252, 412, 444, 796], dtype='i4')
        self.idx2 = np.array([220, 252, 284, 316, 412, 444, 540, 604,
                              668, 700, 796, 828, 924, 988, 1052, 1084,
                              1180, 1212, 1308, 1340, 1372, 1404, 1436, 1468],
                             dtype='i4')

    def calculate(self, system):
        """Calculate area and volume."""
        pos = system.particles.pos
        ar_ring = ConvexHull(pos[self.idx1]).area
        vol_cage = ConvexHull(pos[self.idx2]).volume
        return [ar_ring, vol_cage]


class Position(OrderParameter):
    """A positional order parameter.

    This class defines a very simple order parameter which is just
    the position of a given particle.

    Attributes
    ----------
    index : integer
        This is the index of the atom which will be used, i.e.
        ``system.particles.pos[index]`` will be used.
    dim : integer
        This is the dimension of the coordinate to use.
        0, 1 or 2 for 'x', 'y' or 'z'.
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the position or not.

    """

    def __init__(self, index, dim='x', periodic=False):
        """Initialise the order parameter.

        Parameters
        ----------
        index : int
            This is the index of the atom we will use the position of.
        dim : string
            This select what dimension we should consider,
            it should equal 'x', 'y' or 'z'.
        periodic : boolean, optional
            This determines if periodic boundary conditions should be
            applied to the position.

        """
        txt = 'Position of particle {} (dim: {})'.format(index, dim)
        super().__init__(description=txt, velocity=False)
        self.periodic = periodic
        self.index = index
        self.dim = {'x': 0, 'y': 1, 'z': 2}.get(dim, None)
        if self.dim is None:
            msg = 'Unknown dimension {} requested'.format(dim)
            raise ValueError(msg)

    def calculate(self, system):
        """Calculate the position order parameter.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The object containing the positions.

        Returns
        -------
        out : list of floats
            The position order parameter.

        """
        particles = system.particles
        pos = particles.pos[self.index]
        lamb = pos[self.dim]
        if self.periodic:
            lamb = system.box.pbc_coordinate_dim(lamb, self.dim)
        return [lamb]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the common methods in pyretis.pyvisa.plotting."""
import os
import logging
import unittest
import numpy as np
from matplotlib import pyplot
from matplotlib.lines import Line2D
from mpl_toolkits.mplot3d import Axes3D  # It is needed
from pyretis.pyvisa.plotting import (
    plot_regline,
    plot_int_plane,
    gen_surface,
    _grid_it_up)

logging.disable(logging.CRITICAL)

HERE = os.path.abspath(os.path.dirname(__file__))


class TestMethods(unittest.TestCase):
    """Test some of the methods from pyretis.pyvisa.plotting."""

    def test_plot_regline(self):
        """Test of the regression plotting"""
        _, ax = pyplot.subplots()
        x = [i for i in range(10)]
        y = [i**2 - 1. for i in x]
        # Expected regression line points
        ptm = [x[0]*9. - 13., x[-1]*9. - 13.]
        line = Line2D([x[0], x[-1]], ptm)
        # Regression line
        regline = plot_regline(ax, x, y)
        # Loop over y-values and compare
        for i, j in zip(regline[0].get_ydata(), line.get_ydata()):
            self.assertEqual(i, j)

    def test_plot_int_plane(self):
        """Test if we can make an interface plane in 3d plot"""
        fig = pyplot.figure()
        ax = fig.add_subplot(111, projection='3d')
        # Interface plane specifics
        pos = 1.0
        ymin, ymax, zmin, zmax = 0.0, 1.0, 0.0, 1.0
        visible = False
        _ = plot_int_plane(ax, pos, (ymin, ymax), (zmin, zmax),
                           visible=visible)
        # Nothing to check, only if any errors should arise from plotting
        fig.clear()

    def test_gen_surface(self):
        """Test if we can generate a surface/plot with given settings"""
        # Create data
        x, y, z = [], [], []
        for i in range(-10, 10):
            for j in range(-10, 10):
                x.append(i/10.)
                y.append(j/10.)
                z.append((i/10.)**2 + (j/10.)**2)
        # Create figure and plot+colorbar axes
        fig = pyplot.figure()
        ax = fig.add_subplot(111, projection='3d')
        cbar_ax = fig.add_axes([0.86, 0.1, 0.03, 0.8])
        # Generate surface 3d
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax,
                        res_x=200, res_y=200, method='surface')
        ax.clear()
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax,
                        res_x=200, res_y=200, method='contour')
        ax.clear()
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax,
                        res_x=200, res_y=200, method='contourf')
        ax.clear()
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax,
                        method='scatter')
        ax.clear()

        fig.clear()
        # Generate surface 2d
        ax = fig.add_subplot(111)
        cbar_ax = fig.add_axes([0.1, 0.03, 0.8, 0.1])
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax, dim=2,
                        method='contourf')
        ax.clear()
        _ = gen_surface(x, y, z, fig, ax, cbar_ax=cbar_ax, dim=2,
                        method='contour')
        ax.clear()
        _ = gen_surface(x, y, len(x)*[0.0], fig, ax, cbar_ax=cbar_ax, dim=2,
                        method='scatter')
        ax.clear()

    def test_grid_it_up(self):
        """Test grid method"""
        x = np.array([1, 1, 1, 1, 1, 2, 9])
        y = np.array([1, 2, 2, 4, 4, 1, 9])
        z = np.array([1, 2, 7, 4, 5, 1, 1])

        mins = _grid_it_up([x, y, z], res_x=20, res_y=20, fill='max')
        maxs = _grid_it_up([x, y, z], res_x=20, res_y=20, fill='min')
        self.assertFalse(np.equal(mins, maxs).all())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains common functions for the path density.

It contains some functions that is used to compare and process data,
like matching similar lists or attempt periodic shifts of values.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

find_rst_file (:py:func: `.find_rst_file`)
    Search for a rst-file from a chosen subdirectory.

read_traj_txt_file (:py:func: `.read_traj_txt_file`)
    Read the sequence of files in a trajectory from a traj.txt file.

recalculate_all (:py:func:`.recalculate_all`)
    Recalculate order parameter and new collective variables
    by finding all trajectory files from a simulation.

shift_data (:py:func: `.shift_data`)
    Finds the median value of a given list of floats, and shifts the
    lower half of the data by the median.

try_data_shift (:py:func: `.try_data_shift`
    Takes in two lists of values, x and y, and calculates a linear
    regression and R**2-correlation of the data set. Attempts a shift of
    each data set by their respective median to increase the correlation.

where_from_to (:py:func: `.where_from_to`)
    Check the initial and final steps of a trj in respect to the interfaces.

get_cv_names (:py:func: `.get_cv_names`)
    Outputs a list of the names of the descriptors in the simulation.

recalculate_all (:py:func: `.recalculate_all`)
    Recompute all the order parameters according to the PyRETIS storage scheme
    or for individual files/folders

find_data  (:py:func: `.find_data`)
    Find suitable frames/trajectories to recompute the orderp parameter on.

"""
import os
import timeit

import scipy

from pyretis.initiation.initiate_load import write_order_parameters
from pyretis.inout import print_to_screen, settings
from pyretis.inout.common import create_backup, TRJ_FORMATS
from pyretis.inout.formats.path import PathExtFile
from pyretis.setup.common import create_orderparameter
from pyretis.tools.recalculate_order import recalculate_order

__all__ = ['find_rst_file', 'read_traj_txt_file',
           'shift_data', 'try_data_shift', 'where_from_to',
           'get_cv_names', 'recalculate_all', 'find_data']


def try_data_shift(x, y, fixedx):
    """Check if shifting increases correlation.

    Function that checks if correlation of data increases by shifting
    either sets of values, x or y, or both. Correlation is checked by
    doing a simple linear regression on the different sets of data:
    - x and y , x and yshift, xshift and y, xshift and yshift.
    If linear correlation increases (r-squared value), data sets are
    updated.

    As a precaution, no shift
    is performed on x values if they are of the first order parameter 'op1'.

    Parameters
    ----------
    x, y : list
        Floats, data values
    fixedx : bool
        If True, x is main OP and should not be shifted.

    Returns
    -------
    x, y : list
        Floats, updated (or unchanged) data values
        (If changed, returns x_temp or y_temp or both)

    """
    # The unshifted data
    _, _, r_val, _, _ = scipy.stats.linregress(x, y)
    # The Y-shifted data
    y_temp = shift_data(y)
    _, _, r_y, _, _ = scipy.stats.linregress(x, y_temp)
    yshift = r_y**2 > r_val**2
    # The X-shifted data
    x_temp = shift_data(x)
    _, _, r_x, _, _ = scipy.stats.linregress(x_temp, y)
    xshift = r_x**2 > r_val**2 and r_x**2 > r_y**2
    # Comparing effectiveness of both shifts individually, and combined
    _, _, r_xy, _, _ = scipy.stats.linregress(x_temp, y_temp)
    xyshift = r_xy**2 > r_val**2 and r_xy**2 > r_y**2 and r_xy**2 > r_x**2

    # If first op is op1, don't shift data
    if xyshift and not fixedx:
        return x_temp, y_temp
    if xshift and not fixedx:
        return x_temp, y
    if yshift:
        return x, y_temp
    return x, y


def shift_data(x):
    """Shifts the data under the median.

    Function that takes in a list of data, and shifts all values
    below the median value of the data by the max difference,
    effectively shifting parts of the data periodically in order to
    give clusters for visualization.

    Parameters
    ----------
    x : list
        Floats, data values

    Returns
    -------
    xnorm : list
        Floats where some values are shifted values of x,
        and some are left unchanged.

    """
    xmin, xmax = min(x), max(x)
    xnorm = []
    # The max difference in x-data
    diff_x = xmax - xmin
    # The Median of x-data
    medix = xmin + 0.5 * diff_x

    for i in x:
        if i < medix:
            xnorm.append(i + diff_x)
        else:
            xnorm.append(i)
    return xnorm


def read_traj_txt_file(path):
    """Read a traj.txt file.

    Function which reads a traj.txt file and returns a dict containing
    the name of each file in the trajectory and the sign of their velocity.

    Parameters
    ----------
    path : string
        Path to the traj.txt file.

    Returns
    -------
    files : dict
        Dictionary containing each file in the trajectory and
        the sign of their velocity.

    """
    files = {}
    i = 0
    with PathExtFile(path, 'r') as pfile:
        for block in pfile.load():
            for data in block['data']:
                if data[0] == '0':
                    files[i] = [data[1], data[3]]

                if data[1] != files[i][0]:
                    i += 1
                    files[i] = [data[1], data[3]]
    return files


def find_rst_file(search_dir):
    """Search for rst-files.

    Parameters
    ----------
    search_dir : string
        Path to the .rst file.

    Returns
    -------
    out[0] : string
        Path and name of the .rst file.

    """
    for _ in range(len(search_dir.split('/'))):
        for file_name in os.listdir(search_dir):
            if file_name.endswith('rst'):
                os.chdir(search_dir)
                return search_dir + '/' + file_name
        os.chdir('../')
        search_dir = os.getcwd()
    return search_dir


def where_from_to(trj, int_a, int_b=float('-inf')):
    r"""Detect L∕R starts and L / R / \* ends.

    Given a list of order parameters (a trj), the function
    will try to establish where the path started (L or R or \*)
    and where it ended.
    Note: for the 'REJ' paths, this function results might differ
    from PyRETIS.

    Parameters
    ----------
    trj: numpy array
        The order parameters of the trj.
    int_a: float
        The interface that defines state A.
    int_b: float, optional
        The interface that defines state B. If not given, it is assumed
        that the 0^- ensemble is in use without the 0^- L interface.

    Returns
    -------
    start: string\*1
        The initial position of the trajectory in respect to the
        interfaces given (L eft, R ight or \* for nothing).
    end: string\*1
        The final position of the trajectory in respect to the
        interfaces given (L eft, R ight or \* for nothing).

    """
    start, end = '*', '*'
    int_l = min(int_a, int_b)
    int_t = max(int_a, int_b)
    if trj[0] >= int_t:
        start = 'R'
    if trj[0] < int_l:
        start = 'L'
    if trj[-1] >= int_t:
        end = 'R'
    if trj[-1] < int_l:
        end = 'L'

    return start, end


def get_cv_names(input_settings):
    """Return names of op and cv's.

    Parameters
    ----------
    input_settings: dict
        Dictionary with the settings from the simulations.

    Returns
    -------
    names = list
        List of the names.

    """
    op_names = []
    # Collect names of op and cv's if available
    if 'name' in input_settings['orderparameter'].keys():
        op_names.append(input_settings['orderparameter']['name'])
    if 'collective-variable' in input_settings.keys():
        for c_v in input_settings['collective-variable']:
            if 'name' in c_v.keys():
                op_names.append(c_v['name'])
    return op_names


def recalculate_all(runfolder, iofile, ensemble_names=None, data=None):
    """Recalculate order parameter and collective variables.

    This function performs post-processing by analyzing trajectories
    of old simulations in order to extract data and do new calculations
    and write to a new order.txt file.

    Parameters
    ----------
    runfolder: string
        The path of the execution directory.
    iofile: string
        The input file where the settings are collected.
    ensemble_names: list, optional
        List of ensemble names in the simulation to work with.
    data: string, optional
        If given, the function will check only the single file or
        look only in the given directory

    Returns
    -------
    out : boolean
        True if the recomputation was successful, False otherwise.

    """
    # Gather collective variables and corresponding option from .rst
    io_file = os.path.join(runfolder, iofile)
    assert io_file is not None, 'Input file not given'
    input_settings = settings.parse_settings_file(io_file)
    trj_dict = find_data(runfolder, ensemble_names, data=data)

    print_to_screen('Re-computing the collective variables.', level='message')
    tic = timeit.default_timer()
    if not trj_dict:
        print_to_screen('No data to re-compute from', level='warning')
        return False

    try:
        functions = create_orderparameter(input_settings)
    except (ImportError, ValueError):
        msg = 'Invalid Order Parameter'
        print_to_screen(msg, level='warning')  # pragma: no cover
        return False  # pragma: no cover

    # Create the composite order parameter (Avoid circular reference)
    for _, ens in trj_dict.items():
        if ens.get('main_o') is not None:
            main_order = os.path.join(runfolder, ens['main_o'])
            create_backup(main_order)
        for _, cycles in sorted(ens['traj'].items()):
            here = os.path.dirname(os.path.abspath(cycles['traj'][0]))
            new_o = os.path.join(here, 'order.txt')
            results_dict = []
            for trj in cycles['traj']:
                try:
                    for order_p in recalculate_order(functions, trj, {}):
                        results_dict.append(order_p)  # pragma: no cover
                except (KeyError, AttributeError):
                    print_to_screen(f'File {trj} not valid', level='warning')

            local_order = cycles.get('o_txt', new_o)
            create_backup(local_order)
            write_order_parameters(local_order, results_dict,
                                   cycles.get('header', 'Recalculated data'))

            if ens.get('main_o') is not None:
                os.system(f"cat {local_order} >> {main_order}")

    print_to_screen('# Data successfully recomputed!', level='success')
    print_to_screen(f'# Time spent: {timeit.default_timer() - tic:.2f}s',
                    level='success')
    return True


def find_data(runfolder, ensemble_names=None, data=None):
    """Find the trajectory data used to do post-processing.

    find_traj returns a dict with a structure resembling that
    of the simulation.

    Parameters
    ----------
    runfolder: string, optional
        The path of the execution directory.
    ensemble_names: list, optional
        List of ensemble names in the simulation to work with.
    data: string, optional
        If given, the function will check only the single file or
        look only in the given directory

    Returns
    -------
    trj_dict : dict
        To each key, ensemble_name (e.g. 000, 001, etc)
        the values are: the last accepted trajectories given by the
        `accepted`-key; the generation trajectory or conf files given by
        the `generation`-key, and lastly the dictionary `stored_traj`
        that is given by the `traj`-key. `stored_traj` is split up into
        the dictionaries`traj-acc` and `traj-rej` which have keys for
        all the accepted and rejected cycles respectively,
        where the trajectory files for that cycle is stored.

    """
    trj_dict = {}
    # Specific file/folder
    if data is not None:
        sources = [data] if os.path.isfile(data) else _get_trjs(data)
        if sources:
            trj_dict['000'] = {'traj': {'0': {'traj': sources}}}
        return trj_dict

    # Structured data
    flag_map = {'traj-acc': 'ACC', 'traj-rej': 'REJ'}

    if ensemble_names is None:
        ensemble_names = [i.name for i in os.scandir(runfolder) if
                          (i.is_dir() and i.name.isdigit())]

    for ens_name in sorted(ensemble_names):
        order_txt = os.path.join(os.path.abspath(runfolder), ens_name,
                                 'order.txt')
        trj_dict[ens_name] = {'main_o': order_txt, 'traj': {}}
        for trj_type in ['traj-acc', 'traj-rej']:
            here = os.path.join(os.path.abspath(runfolder),
                                ens_name, 'traj', trj_type)
            if not os.path.isdir(here):
                continue
            for cycle in sorted([i.name for i in os.scandir(here) if
                                i.name.isdigit()]):
                loc = os.path.join(here, cycle, 'traj')
                o_txt = os.path.join(here, cycle, 'order.txt')
                header = f'# Cycle: {cycle},' \
                         f' status: {flag_map[trj_type]}'
                if os.path.exists(o_txt):
                    with open(o_txt, 'r', encoding='utf-8') as file_in:
                        header = file_in.readline().replace('\n', '')

                trj_dict[ens_name]['traj'][cycle] = {'header': header,
                                                     'o_txt': o_txt,
                                                     'traj': _get_trjs(loc)}
    return trj_dict


def _get_trjs(runfolder='.'):
    """Find the trajectory files.

    Parameters
    ----------
    runfolder : string, optional
        the location of the main simulation folder.

    Returns
    -------
    full_name_trj : list
        The trajectory files contained in the folder.

    """
    trj = [i.name for i in os.scandir(runfolder) if i.name[-4:] in TRJ_FORMATS]
    full_name_trj = [os.path.join(runfolder, i) for i in trj]
    return full_name_trj



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test of the pairpotential module."""
import logging
import unittest
import numpy as np
from pyretis.forcefield.potentials.pairpotentials.pairpotential import (
    mixing_parameters,
    generate_pair_interactions,
    _check_pair_parameters,
)
logging.disable(logging.CRITICAL)


class TestMixingParameters(unittest.TestCase):
    """Test that we can set up for mixing."""

    def test_mixing_geometric(self):
        """Test the geometric mixing."""
        epsilon_i, epsilon_j = 2.0, 3.0
        sigma_i, sigma_j = 5.0, 6.0
        rcut_i, rcut_j = 10.0, 11.0
        mixed = mixing_parameters([epsilon_i, epsilon_j],
                                  [sigma_i, sigma_j],
                                  [rcut_i, rcut_j],
                                  mixing='geometric')
        for i, j in zip(mixed, ((epsilon_i * epsilon_j),
                                (sigma_i * sigma_j),
                                (rcut_i * rcut_j))):
            self.assertEqual(i, np.sqrt(j))

    def test_mixing_arithmetic(self):
        """Test arithmetic mixing."""
        epsilon_i, epsilon_j = 2.0, 3.0
        sigma_i, sigma_j = 5.0, 6.0
        rcut_i, rcut_j = 10.0, 11.0
        correct = (np.sqrt(epsilon_i * epsilon_j),
                   0.5 * (sigma_i + sigma_j),
                   0.5 * (rcut_i + rcut_j))
        mixed = mixing_parameters([epsilon_i, epsilon_j],
                                  [sigma_i, sigma_j],
                                  [rcut_i, rcut_j],
                                  mixing='arithmetic')
        for i, j in zip(mixed, correct):
            self.assertEqual(i, j)

    def test_mixing_sixthpower(self):
        """Test 6th power mixing."""
        epsilon_i, epsilon_j = 2.0, 3.0
        sigma_i, sigma_j = 5.0, 6.0
        rcut_i, rcut_j = 10.0, 11.0
        si3 = sigma_i**3
        si6 = si3**2
        sj3 = sigma_j**3
        sj6 = sj3**2
        avgs6 = 0.5 * (si6 + sj6)
        mixed = mixing_parameters([epsilon_i, epsilon_j],
                                  [sigma_i, sigma_j],
                                  [rcut_i, rcut_j],
                                  mixing='sixthpower')
        correct = (np.sqrt(epsilon_i * epsilon_j) * si3 * sj3 / avgs6,
                   avgs6**(1.0 / 6.0),
                   (0.5 * (rcut_i**6 + rcut_j**6))**(1.0 / 6.0))
        for i, j in zip(mixed, correct):
            self.assertEqual(i, j)

    def test_mixing_missing(self):
        """Test mixing when we supply something not recognized."""
        logging.disable(logging.INFO)
        mod = 'pyretis.forcefield.potentials.pairpotentials.pairpotential'
        with self.assertLogs(mod, level='WARNING'):
            mixed = mixing_parameters([11., 14], [12., 15], [13., 16.],
                                      mixing='Life On Mars?')
        logging.disable(logging.CRITICAL)
        for i in mixed:
            self.assertEqual(i, 0.0)


class TestGeneratePairInteraction(unittest.TestCase):
    """Test generation of pair parameters."""

    def test_generate_full(self):
        """Test generation of full parameters."""
        parameters = {0: {'sigma': 1.0, 'epsilon': 1.5, 'rcut': 2.},
                      1: {'sigma': 2.0, 'epsilon': 2.5, 'rcut': 3.},
                      2: {'sigma': 3.0, 'epsilon': 3.5, 'rcut': 4.}}
        mix = generate_pair_interactions(parameters, 'arithmetic')
        correct_mix = {
            (0, 0): {'rcut': 2.0, 'epsilon': 1.5, 'sigma': 1.0},
            (2, 2): {'rcut': 4.0, 'epsilon': 3.5, 'sigma': 3.0},
            (1, 1): {'rcut': 3.0, 'epsilon': 2.5, 'sigma': 2.0},
            (0, 1): {'rcut': 2.5, 'epsilon': 1.9364916731037085, 'sigma': 1.5},
            (0, 2): {'rcut': 3.0, 'epsilon': 2.2912878474779199, 'sigma': 2.0},
            (1, 2): {'rcut': 3.5, 'epsilon': 2.9580398915498081, 'sigma': 2.5},
        }
        for key, val in correct_mix.items():
            for key2, val2 in val.items():
                self.assertEqual(val2, mix[key][key2])
        for key1, key2 in zip(((1, 0), (2, 0), (2, 1)),
                              ((0, 1), (0, 2), (1, 2))):
            for key, val in mix[key1].items():
                self.assertEqual(val, mix[key2][key])

    def test_generate_pair(self):
        """Testing generate where we specify some cross interactions."""
        parameters = [
            {0: {'sigma': 1.0, 'epsilon': 1.5, 'rcut': 2.},
             1: {'sigma': 2.0, 'epsilon': 2.5, 'rcut': 3.},
             (0, 1): {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 100.}},
            {0: {'sigma': 1.0, 'epsilon': 1.5, 'rcut': 2.},
             1: {'sigma': 2.0, 'epsilon': 2.5, 'rcut': 3.},
             (1, 0): {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 100.}},
        ]
        for param in parameters:
            mix = generate_pair_interactions(param, 'arithmetic')
            correct_mix = {
                (0, 0): param[0],
                (1, 1): param[1],
            }
            if (0, 1) in param:
                correct_mix[(0, 1)] = param[(0, 1)]
                correct_mix[(1, 0)] = param[(0, 1)]
            elif (1, 0) in param:
                correct_mix[(0, 1)] = param[(1, 0)]
                correct_mix[(1, 0)] = param[(1, 0)]
            for key, val in correct_mix.items():
                for key2, val2 in val.items():
                    self.assertEqual(val2, mix[key][key2])

    def test_check_pair_params(self):
        """Test the check pair params."""
        parameters = {
            0: {'sigma': 1.0, 'epsilon': 1.5, 'rcut': 2.},
            1: {'sigma': 2.0, 'rcut': 3.}
        }
        self.assertFalse('epsilon' in parameters[1])
        _check_pair_parameters(parameters)
        self.assertTrue('epsilon' in parameters[1])
        self.assertAlmostEqual(parameters[1]['epsilon'], 0.0)
        # Missing rcut:
        parameters = {0: {'sigma': 1.0, 'epsilon': 1.5, 'factor': 123.}}
        self.assertFalse('rcut' in parameters[0])
        _check_pair_parameters(parameters)
        self.assertTrue('rcut' in parameters[0])
        self.assertAlmostEqual(parameters[0]['rcut'], 123.)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two pandas dataframe files."""
import sys
import pandas as pd


def main():
    """Compare two pandas files.

    Returns
    -------
    out : int
        0 if the comparison was successful, 1 otherwise.

    """
    df1 = pd.read_hdf('pyvisa_compressed_data.hdf5', key='data')
    df2 = pd.read_hdf('results/pyvisa_compressed_data.hdf5', key='data')

    # Compare pandas DataFrames _data_

    # Remove NaN, since NaN != NaN
    for ens1, ens2 in zip(df1, df2):
        for trj1, trj2 in zip(ens1, ens2):
            ens1[trj1].frames.notna()
            ens2[trj2].frames.notna()

            # Now compare the remaining
            if not ens1[trj1].frames.equals(ens1[trj2].frames):
                print(ens1[trj1].frames)
                print('The files contain different data')
                return 1

    # Compare pandas DataFrames _settings_
    df1 = pd.read_hdf('pyvisa_compressed_data.hdf5', key='infos')
    df2 = pd.read_hdf('results/pyvisa_compressed_data.hdf5', key='infos')
    if not df1.equals(df2):
        print('The files contain different settings')
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())
